#!/usr/bin/env python3

# Copyright © 2023-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    Type, Union, Optional, Literal, List
)
from pyaes import AESModeOfOperationECB

import scrypt
import unicodedata
import os

from .cryptocurrencies import (
    ICryptocurrency, Bitcoin
)
from .libs.base58 import (
    encode, check_encode, decode, check_decode, ensure_string
)
from .secp256k1 import (
    PublicKey, PrivateKey
)
from .p2pkh_address import P2PKHAddress
from .crypto import (
    double_sha256, get_checksum
)
from .const import (
    NP,
    MAGIC_LOT_AND_SEQUENCE,
    MAGIC_NO_LOT_AND_SEQUENCE,
    NO_EC_MULTIPLIED_WIF_FLAG,
    NO_EC_MULTIPLIED_WIF_COMPRESSED_FLAG,
    NO_EC_MULTIPLIED_PRIVATE_KEY_PREFIX,
    MAGIC_LOT_AND_SEQUENCE_UNCOMPRESSED_FLAG,
    MAGIC_LOT_AND_SEQUENCE_COMPRESSED_FLAG,
    MAGIC_NO_LOT_AND_SEQUENCE_UNCOMPRESSED_FLAG,
    MAGIC_NO_LOT_AND_SEQUENCE_COMPRESSED_FLAG,
    EC_MULTIPLIED_PRIVATE_KEY_PREFIX,
    CONFIRMATION_CODE_PREFIX,
    WIF_TYPES,
    FLAGS
)
from .exceptions import (
    Error, CryptocurrencyError, NetworkError, Secp256k1Error, WIFError, PassphraseError
)
from .utils import (
    integer_to_bytes, bytes_to_integer, bytes_to_string, get_bytes
)
from .wif import (
    private_key_to_wif, wif_to_private_key, get_wif_type
)


class BIP38:
    """
    A class for BIP38 encryption and decryption of Wallet Import Format (WIF) keys.

    :param cryptocurrency: The cryptocurrency class to use (e.g., Bitcoin).
    :type cryptocurrency: Type[ICryptocurrency]
    :param network: The network for the WIF key (e.g., 'mainnet' or 'testnet'). Defaults to 'mainnet'.
    :type network: str
    """

    cryptocurrency: Type[ICryptocurrency]
    networks: List[str]
    network: str
    alphabet: str

    def __init__(
        self, cryptocurrency: Type[ICryptocurrency], network: str = "mainnet"
    ) -> None:

        if not issubclass(cryptocurrency, ICryptocurrency):
            raise CryptocurrencyError(
                "Invalid cryptocurrency sub-class", expected=Type[ICryptocurrency], got=type(cryptocurrency)
            )
        self.cryptocurrency, self.networks = (
            cryptocurrency, list(cryptocurrency.NETWORKS.keys())
        )
        if network not in self.networks:
            raise NetworkError("Wrong network type", expected=self.networks, got=type(network))
        self.network, self.alphabet = network, (
            self.cryptocurrency.ALPHABET
            if self.cryptocurrency.ALPHABET else
            Bitcoin.ALPHABET
        )

    @classmethod
    def intermediate_code(
        cls,
        passphrase: str,
        lot: Optional[int] = None,
        sequence: Optional[int] = None,
        owner_salt: Optional[Union[str, bytes]] = None,
        N: int = 16384, r: int = 8, p: int = 8
    ) -> str:
        """
        Generates an intermediate passphrase.

        :param passphrase: The passphrase or password.
        :type passphrase: str
        :param lot: Optional lot number (100000-999999).
        :type lot: Optional[int]
        :param sequence: Optional sequence number (0-4095).
        :type sequence: Optional[int]
        :param owner_salt: Optional owner salt. (default: ``random 8 bytes``)
        :type owner_salt: Optional[str, bytes]
        :param N: CPU/memory cost parameter (must be a power of two > 1). Higher values increase security but require more resources. (default: ``16384``)
        :type N: int
        :param r: Block size parameter. Controls memory usage. Must be > 0. (default: ``8``)
        :type r: int
        :param p: Parallelization parameter. Defines how many threads can run in parallel. Must be > 0. (default: ``8``)
        :type p: int

        :returns: The intermediate passphrase.
        :rtype: str

        >>> from bip38 import BIP38
        >>> from bip38.cryptocurrencies import Bitcoin
        >>> bip38: BIP38 = BIP38(cryptocurrency=Bitcoin, network="mainnet")
        >>> bip38.intermediate_code(passphrase="TestingOneTwoThree")
        'passphraseqVKbgU4mWMakKGgCtaeVWoETQdzMBy5696bG2w7ckVBeQmoLhMF9vLaxhmzhT3'
        >>> bip38.intermediate_code(passphrase="TestingOneTwoThree", lot=199999, sequence=1, owner_salt="75ed1cdeb254cb38")
        'passphraseb7ruSN4At4Rb8hPTNcAVezfsjonvUs4Qo3xSp1fBFsFPvVGSbpP2WTJMhw3mVZ'
        """

        owner_salt: bytes = get_bytes(owner_salt) if owner_salt else os.urandom(8)
        if len(owner_salt) not in [4, 8]:
            raise Error("Invalid owner salt length", expected=[8, 16], got=len(bytes_to_string(owner_salt)))
        if len(owner_salt) == 4 and (lot is None or sequence is None):
            raise Error("Invalid owner salt length for non lot/sequence", expected=16, got=len(bytes_to_string(owner_salt)))
        if (lot and sequence is None) or (lot is None and sequence):
            raise Error("Both lot & sequence are required", detail=f"got: (lot {lot}) & (sequence {sequence})")

        if lot and sequence:
            lot, sequence = int(lot), int(sequence)
            if not 100000 <= lot <= 999999:
                raise Error("Invalid lot", expected="100000 <= lot <= 999999", got=lot)
            if not 0 <= sequence <= 4095:
                raise Error("Invalid sequence", expected="0 <= sequence <= 4095", got=sequence)

            pre_factor: bytes = scrypt.hash(unicodedata.normalize("NFC", passphrase), owner_salt[:4], N=N, r=r, p=p, buflen=32)
            owner_entropy: bytes = owner_salt[:4] + integer_to_bytes((lot * 4096 + sequence), 4)
            pass_factor: bytes = double_sha256(pre_factor + owner_entropy)
            magic: bytes = integer_to_bytes(MAGIC_LOT_AND_SEQUENCE)
        else:
            pass_factor: bytes = scrypt.hash(unicodedata.normalize("NFC", passphrase), owner_salt, N=N, r=r, p=p, buflen=32)
            magic: bytes = integer_to_bytes(MAGIC_NO_LOT_AND_SEQUENCE)
            owner_entropy: bytes = owner_salt

        pass_point: PublicKey = PrivateKey.from_bytes(
            private_key=pass_factor
        ).public_key()
        return ensure_string(check_encode(
            magic + owner_entropy + pass_point.raw_compressed()
        ))

    def encrypt(
        self, wif: str, passphrase: str, network: Optional[str] = None, N: int = 16384, r: int = 8, p: int = 8
    ) -> str:
        """
        Encrypts a Wallet Import Format (WIF) key using a passphrase with BIP38 encryption.

        :param wif: The WIF key to be encrypted.
        :type wif: str
        :param passphrase: The passphrase for encryption.
        :type passphrase: str
        :param network: Optional network for encryption. Defaults to the class's network if not provided.
        :type network: Optional[str]
        :param N: CPU/memory cost parameter (must be a power of two > 1). Higher values increase security but require more resources. (default: ``16384``)
        :type N: int
        :param r: Block size parameter. Controls memory usage. Must be > 0. (default: ``8``)
        :type r: int
        :param p: Parallelization parameter. Defines how many threads can run in parallel. Must be > 0. (default: ``8``)
        :type p: int

        :returns: Encrypted WIF key as a string.
        :rtype: str

        >>> from bip38 import BIP38
        >>> from bip38.cryptocurrencies import Bitcoin
        >>> bip38: BIP38 = BIP38(cryptocurrency=Bitcoin, network="mainnet")
        >>> bip38.encrypt(wif="5KN7MzqK5wt2TP1fQCYyHBtDrXdJuXbUzm4A9rKAteGu3Qi5CVR", passphrase="TestingOneTwoThree")
        '6PRVWUbkzzsbcVac2qwfssoUJAN1Xhrg6bNk8J7Nzm5H7kxEbn2Nh2ZoGg'
        >>> bip38.encrypt(wif="L44B5gGEpqEDRS9vVPz7QT35jcBG2r3CZwSwQ4fCewXAhAhqGVpP", passphrase="TestingOneTwoThree")
        '6PYNKZ1EAgYgmQfmNVamxyXVWHzK5s6DGhwP4J5o44cvXdoY7sRzhtpUeo'
        >>> bip38.encrypt(wif="938jwjergAxARSWx2YSt9nSBWBz24h8gLhv7EUfgEP1wpMLg6iX", passphrase="TestingOneTwoThree", network="testnet")
        '6PRL8jj6dLQjBBJjHMdUKLSNLEpjTyAfmt8GnCnfT87NeQ2BU5eAW1tcsS'
        >>> bip38.encrypt(wif="cURAYbG6FtvUasdBsooEmmY9MqUfhJ8tdybQWV7iA4BAwunCT2Fu", passphrase="TestingOneTwoThree", network="testnet")
        '6PYVB5nHnumbUua1UmsAMPHWHa76Ci48MY79aKYnpKmwxeGqHU2XpXtKvo'
        """

        network: str = (
            network if network else self.network
        )
        if network not in self.networks:
            raise NetworkError("Wrong network type", expected=self.networks, got=type(network))
        wif_type: str = get_wif_type(
            wif=wif, cryptocurrency=self.cryptocurrency, network=network
        )
        if wif_type == "wif":
            flag: bytes = integer_to_bytes(NO_EC_MULTIPLIED_WIF_FLAG)
            private_key: PrivateKey = PrivateKey.from_bytes(get_bytes(wif_to_private_key(
                wif=wif, cryptocurrency=self.cryptocurrency, network=network
            )))
            public_key_type: str = "uncompressed"
        else:
            flag: bytes = integer_to_bytes(NO_EC_MULTIPLIED_WIF_COMPRESSED_FLAG)
            private_key: PrivateKey = PrivateKey.from_bytes(get_bytes(wif_to_private_key(
                wif=wif, cryptocurrency=self.cryptocurrency, network=network
            )))
            public_key_type: str = "compressed"

        address: str = P2PKHAddress.encode(
            public_key=private_key.public_key(),
            address_prefix=self.cryptocurrency.NETWORKS[network]["address_prefix"],
            public_key_type=public_key_type
        )
        address_hash: bytes = get_checksum(get_bytes(address, unhexlify=False))
        key: bytes = scrypt.hash(unicodedata.normalize("NFC", passphrase), address_hash, N=N, r=r, p=p)
        derived_half_1, derived_half_2 = key[0:32], key[32:64]

        aes: AESModeOfOperationECB = AESModeOfOperationECB(derived_half_2)
        encrypted_half_1: bytes = aes.encrypt(integer_to_bytes(
            bytes_to_integer(private_key.raw()[0:16]) ^ bytes_to_integer(derived_half_1[0:16])
        ))
        encrypted_half_2: bytes = aes.encrypt(integer_to_bytes(
            bytes_to_integer(private_key.raw()[16:32]) ^ bytes_to_integer(derived_half_1[16:32])
        ))

        encrypted_private_key: bytes = (
            integer_to_bytes(
                NO_EC_MULTIPLIED_PRIVATE_KEY_PREFIX
            ) + flag + address_hash + encrypted_half_1 + encrypted_half_2
        )
        return ensure_string(encode(
            encrypted_private_key + get_checksum(encrypted_private_key)
        ))

    def create_new_encrypted_wif(
        self,
        intermediate_passphrase: str,
        wif_type: str = "wif",
        seed: Optional[Union[str, bytes]] = None,
        network: Optional[str] = None
    ) -> dict:
        """
        Creates a new encrypted WIF (Wallet Import Format).

        :param intermediate_passphrase: The intermediate passphrase.
        :type intermediate_passphrase: str
        :param wif_type: The WIF type, either ``wif`` or ``wif-compressed``. (default is ``wif``)
        :type wif_type: str
        :param seed: Optional seed (default: random 24 bytes).
        :type seed: Optional[str, bytes]
        :param network: Optional network for encryption. Defaults to the class's network if not provided.
        :type network: Optional[str]

        :returns: A dictionary containing the encrypted WIF.
        :rtype: dict

        >>> from bip38 import BIP38
        >>> from bip38.cryptocurrencies import Bitcoin
        >>> bip38: BIP38 = BIP38(cryptocurrency=Bitcoin, network="testnet")
        >>> bip38.create_new_encrypted_wif(intermediate_passphrase="passphraseb7ruSN4At4Rb8hPTNcAVezfsjonvUs4Qo3xSp1fBFsFPvVGSbpP2WTJMhw3mVZ", wif_type="wif")
        {'encrypted_wif': '6PgMqfwt1nqJeUkCTRVf4TV6FJoqA8GFCGaj6RsTW1t3XgQNUDfKW9u9Px', 'confirmation_code': 'cfrm38V8ZR17HkU8Xdu8RunWf8CZauXphNQa5HFJd4cxEYckXHS6tfo9M73yL3FPmWv1xqBQsgG', 'public_key': '04fdfbd938b4bb220c11fc1c22e87c5306d105130dd05d7ece4013aa1d2382f3a2a8673fd7b4b2f55c48cd0ebda7d88089783b3394210a0853159803b5eb99097e', 'seed': '0a89d39d0af0f0d7abc655baa0e399f8ddcd372bb9aaebce', 'public_key_type': 'uncompressed', 'address': 'myRfjUS74ab2ZbEbQuiWNzPHb5fSYuFvm4'}
        >>> bip38.create_new_encrypted_wif(intermediate_passphrase="passphraseb7ruSN4At4Rb8hPTNcAVezfsjonvUs4Qo3xSp1fBFsFPvVGSbpP2WTJMhw3mVZ", wif_type="wif-compressed", seed="99241d58245c883896f80843d2846672d7312e6195ca1a6c")
        {'encrypted_wif': '6PoH364JVeoBPsJBveXCwfWpX2H82N5qiAervtynak7r7dzZF2TBFxZAXE', 'confirmation_code': 'cfrm38VX6jSx7C3nbxWCVEdGna7JMpm57zHdbuofbpCA9EiN57aEt4s5fh9k19b5cTmyZ5jMkE2', 'public_key': '02100bb0440ff4106a1743750813271e66a7017431e92921e520319f537c7357c1', 'seed': '99241d58245c883896f80843d2846672d7312e6195ca1a6c', 'public_key_type': 'compressed', 'address': 'mjurfzLk2ryxCyfm4nMk5qarvNRhbNCtK8'}
        """

        network: str = (
            network if network else self.network
        )
        if network not in self.networks:
            raise NetworkError("Wrong network type", expected=self.networks, got=type(network))
        seed_b: bytes = get_bytes(seed) if seed else os.urandom(24)
        try:
            intermediate_decode: bytes = check_decode(intermediate_passphrase)
        except ValueError:
            raise PassphraseError("Invalid intermediate passphrase")
        if len(intermediate_decode) != 49:
            raise PassphraseError("Invalid intermediate passphrase length", expected=49, got=len(intermediate_decode))

        magic: bytes = intermediate_decode[:8]
        owner_entropy: bytes = intermediate_decode[8:16]
        pass_point: bytes = intermediate_decode[16:]

        if wif_type == "wif":
            public_key_type = "uncompressed"
        elif wif_type == "wif-compressed":
            public_key_type = "compressed"
        else:
            raise WIFError("Invalid wif type", expected=WIF_TYPES, got=wif_type)

        if magic == integer_to_bytes(MAGIC_LOT_AND_SEQUENCE):
            if public_key_type == "uncompressed":
                flag: bytes = integer_to_bytes(MAGIC_LOT_AND_SEQUENCE_UNCOMPRESSED_FLAG)
            else:
                flag: bytes = integer_to_bytes(MAGIC_LOT_AND_SEQUENCE_COMPRESSED_FLAG)
        elif magic == integer_to_bytes(MAGIC_NO_LOT_AND_SEQUENCE):
            if public_key_type == "uncompressed":
                flag: bytes = integer_to_bytes(MAGIC_NO_LOT_AND_SEQUENCE_UNCOMPRESSED_FLAG)
            else:
                flag: bytes = integer_to_bytes(MAGIC_NO_LOT_AND_SEQUENCE_COMPRESSED_FLAG)
        else:
            raise Error(
                "Invalid magic", expected=[
                    bytes_to_string(integer_to_bytes(MAGIC_LOT_AND_SEQUENCE)),
                    bytes_to_string(integer_to_bytes(MAGIC_NO_LOT_AND_SEQUENCE))
                ], got=bytes_to_string(magic)
            )

        factor_b: bytes = double_sha256(seed_b)
        if not 0 < bytes_to_integer(factor_b) < NP:
            raise Error("Invalid EC encrypted WIF (Wallet Import Format)")

        public_key: PublicKey = PublicKey.from_point(
            PublicKey.from_bytes(pass_point).point() * bytes_to_integer(factor_b)
        )
        address: str = P2PKHAddress.encode(
            public_key=public_key,
            address_prefix=self.cryptocurrency.NETWORKS[network]["address_prefix"],
            public_key_type=public_key_type
        )
        address_hash: bytes = get_checksum(get_bytes(address, unhexlify=False))
        salt: bytes = address_hash + owner_entropy
        scrypt_hash: bytes = scrypt.hash(pass_point, salt, 1024, 1, 1, 64)
        derived_half_1, derived_half_2, key = scrypt_hash[:16], scrypt_hash[16:32], scrypt_hash[32:]

        aes: AESModeOfOperationECB = AESModeOfOperationECB(key)
        encrypted_half_1: bytes = aes.encrypt(integer_to_bytes(
            bytes_to_integer(seed_b[:16]) ^ bytes_to_integer(derived_half_1)
        ))
        encrypted_half_2: bytes = aes.encrypt(integer_to_bytes(
            bytes_to_integer(encrypted_half_1[8:] + seed_b[16:]) ^ bytes_to_integer(derived_half_2)
        ))
        encrypted_wif: str = ensure_string(check_encode((
            integer_to_bytes(EC_MULTIPLIED_PRIVATE_KEY_PREFIX) +
            flag + address_hash + owner_entropy + encrypted_half_1[:8] + encrypted_half_2
        )))

        point_b: bytes = PrivateKey.from_bytes(factor_b).public_key().raw_compressed()
        point_b_prefix: bytes = integer_to_bytes(
            (bytes_to_integer(scrypt_hash[63:]) & 1) ^ bytes_to_integer(point_b[:1])
        )
        point_b_half_1: bytes = aes.encrypt(integer_to_bytes(
            bytes_to_integer(point_b[1:17]) ^ bytes_to_integer(derived_half_1)
        ))
        point_b_half_2: bytes = aes.encrypt(integer_to_bytes(
            bytes_to_integer(point_b[17:]) ^ bytes_to_integer(derived_half_2)
        ))
        encrypted_point_b: bytes = (
            point_b_prefix + point_b_half_1 + point_b_half_2
        )
        confirmation_code: str = ensure_string(check_encode((
            integer_to_bytes(CONFIRMATION_CODE_PREFIX) + flag + address_hash + owner_entropy + encrypted_point_b
        )))

        return dict(
            encrypted_wif=encrypted_wif,
            confirmation_code=confirmation_code,
            public_key=bytes_to_string(
                public_key.raw(public_key_type=public_key_type)
            ),
            seed=bytes_to_string(seed_b),
            public_key_type=public_key_type,
            address=address
        )

    def confirm_code(
        self, passphrase: str, confirmation_code: str, network: Optional[str] = None, detail: bool = False, N: int = 16384, r: int = 8, p: int = 8
    ) -> Union[str, dict]:
        """
        Confirms the passphrase using a confirmation code.

        :param passphrase: The passphrase or password.
        :type passphrase: str
        :param confirmation_code: The confirmation code.
        :type confirmation_code: str
        :param network: Optional network for encryption. Defaults to the class's network if not provided.
        :type network: Optional[str]
        :param detail: Whether to return detailed info. (default: ``False``)
        :type detail: bool
        :param N: CPU/memory cost parameter (must be a power of two > 1). Higher values increase security but require more resources. (default: ``16384``)
        :type N: int
        :param r: Block size parameter. Controls memory usage. Must be > 0. (default: ``8``)
        :type r: int
        :param p: Parallelization parameter. Defines how many threads can run in parallel. Must be > 0. (default: ``8``)
        :type p: int

        :returns: Confirmation result as a string or detailed info as a dictionary.
        :rtype: Union[str, dict]

        >>> from bip38 import BIP38
        >>> from bip38.cryptocurrencies import Bitcoin
        >>> bip38: BIP38 = BIP38(cryptocurrency=Bitcoin, network="testnet")
        >>> bip38.confirm_code(passphrase="TestingOneTwoThree", confirmation_code="cfrm38V8ZQSdCuzcrYYKGNXVwbHgdjsUEfAbFGoEUouB4YEKaXVdFiMcBWai1Exdu8jN7DcoKtM")
        'mwW38M23zvDmhbsTdnVFzw3bVnueDhrKec'
        >>> bip38.confirm_code(passphrase="TestingOneTwoThree", confirmation_code="cfrm38V8Foq3WpRPMXJD34SF6pGT6ht5ihYMWWMbezkzHgPpA1jVkfbTHwQzvuSA4ReF86PHZJY", network="mainnet")
        '1JbyXoVN4hXWirGB265q9VE4pQ6qbY6kmr'
        >>> bip38.confirm_code(passphrase="TestingOneTwoThree", confirmation_code="cfrm38VXL5T6qVke13sHUWtEjibAkK1RquBqMXb2azCv1Zna6JKvBhD1Gf2b15wBj7UPv2BQnf6", network="mainnet", detail=True)
        {'public_key': '02100bb0440ff4106a1743750813271e66a7017431e92921e520319f537c7357c1', 'public_key_type': 'compressed', 'address': '15PuNwFmDqYhRsC9MDPNFvNY4Npzibm67c', 'lot': 199999, 'sequence': 1}
        """

        network: str = (
            network if network else self.network
        )
        if network not in self.networks:
            raise NetworkError("Wrong network type", expected=self.networks, got=type(network))
        try:
            confirmation_code_decode: bytes = check_decode(confirmation_code)
        except ValueError:
            raise Error("Invalid confirmation code")
        if len(confirmation_code_decode) != 51:
            raise Error("Invalid confirmation code length", expected=102, got=len(confirmation_code_decode))

        prefix_length: int = len(integer_to_bytes(CONFIRMATION_CODE_PREFIX))
        prefix_got: bytes = confirmation_code_decode[:prefix_length]
        if integer_to_bytes(CONFIRMATION_CODE_PREFIX) != prefix_got:
            raise Error("Invalid confirmation code prefix", expected=prefix_length, got=prefix_got)

        flag: bytes = confirmation_code_decode[5:6]
        address_hash: bytes = confirmation_code_decode[6:10]
        owner_entropy: bytes = confirmation_code_decode[10:18]
        encrypted_point_b: bytes = confirmation_code_decode[18:]

        lot_and_sequence: Optional[bytes] = None
        if bytes_to_integer(flag) in FLAGS["lot_and_sequence"]:
            owner_salt: bytes = owner_entropy[:4]
            lot_and_sequence = owner_entropy[4:]
        else:
            owner_salt: bytes = owner_entropy

        pass_factor: bytes = scrypt.hash(unicodedata.normalize("NFC", passphrase), owner_salt, N=N, r=r, p=p, buflen=32)
        if lot_and_sequence:
            pass_factor: bytes = double_sha256(pass_factor + owner_entropy)
        if bytes_to_integer(pass_factor) == 0 or bytes_to_integer(pass_factor) >= NP:
            raise Error("Invalid EC encrypted WIF (Wallet Import Format)")

        pass_point: bytes = PrivateKey.from_bytes(pass_factor).public_key().raw_compressed()
        salt: bytes = address_hash + owner_entropy
        scrypt_hash: bytes = scrypt.hash(pass_point, salt, 1024, 1, 1, 64)
        derived_half_1, derived_half_2, key = encrypted_point_b[1:17], encrypted_point_b[17:], scrypt_hash[32:]

        aes: AESModeOfOperationECB = AESModeOfOperationECB(key)
        point_b_half_1: bytes = integer_to_bytes(
            bytes_to_integer(aes.decrypt(derived_half_1)) ^ bytes_to_integer(scrypt_hash[:16])
        )
        point_b_half_2: bytes = integer_to_bytes(
            bytes_to_integer(aes.decrypt(derived_half_2)) ^ bytes_to_integer(scrypt_hash[16:32])
        )
        point_b_prefix: bytes = integer_to_bytes(
            bytes_to_integer(encrypted_point_b[:1]) ^ (bytes_to_integer(scrypt_hash[63:]) & 1)
        )
        point_b: bytes = (
            point_b_prefix + point_b_half_1 + point_b_half_2
        )

        try:
            public_key: PublicKey = PublicKey.from_point(
                PublicKey.from_bytes(point_b).point() * bytes_to_integer(pass_factor)
            )
            public_key_type: str = "uncompressed"
            if bytes_to_integer(flag) in FLAGS["compression"]:
                public_key_type: str = "compressed"
        except Secp256k1Error:
            raise PassphraseError("Incorrect passphrase")

        address: str = P2PKHAddress.encode(
            public_key=public_key,
            address_prefix=self.cryptocurrency.NETWORKS[network]["address_prefix"],
            public_key_type=public_key_type
        )
        if get_checksum(get_bytes(address, unhexlify=False)) == address_hash:
            lot: Optional[int] = None
            sequence: Optional[int] = None
            if detail:
                if lot_and_sequence:
                    sequence: int = bytes_to_integer(lot_and_sequence) % 4096
                    lot: int = (bytes_to_integer(lot_and_sequence) - sequence) // 4096
                return dict(
                    public_key=bytes_to_string(public_key.raw(public_key_type=public_key_type)),
                    public_key_type=public_key_type,
                    address=address,
                    lot=lot,
                    sequence=sequence
                )
            return address
        raise PassphraseError("Incorrect passphrase")

    def decrypt(
        self, encrypted_wif: str, passphrase: str, network: Optional[str] = None, detail: bool = False, N: int = 16384, r: int = 8, p: int = 8
    ) -> Union[str, dict]:
        """
        Decrypts an encrypted WIF (Wallet Import Format) using a passphrase.

        :param encrypted_wif: The encrypted WIF.
        :type encrypted_wif: str
        :param passphrase: The passphrase or password.
        :type passphrase: str
        :param network: Optional network for encryption. Defaults to the class's network if not provided.
        :type network: Optional[str]
        :param detail: Whether to return detailed info. (default: ``False``)
        :type detail: bool
        :param N: CPU/memory cost parameter (must be a power of two > 1). Higher values increase security but require more resources. (default: ``16384``)
        :type N: int
        :param r: Block size parameter. Controls memory usage. Must be > 0. (default: ``8``)
        :type r: int
        :param p: Parallelization parameter. Defines how many threads can run in parallel. Must be > 0. (default: ``8``)
        :type p: int

        :returns: The decrypted WIF or detailed private key info.
        :rtype: Union[str, dict]

        >>> from bip38 import BIP38
        >>> from bip38.cryptocurrencies import Bitcoin
        >>> bip38: BIP38 = BIP38(cryptocurrency=Bitcoin, network="mainnet")
        >>> bip38.decrypt(encrypted_wif="6PRVWUbkzzsbcVac2qwfssoUJAN1Xhrg6bNk8J7Nzm5H7kxEbn2Nh2ZoGg", passphrase="TestingOneTwoThree")
        '5KN7MzqK5wt2TP1fQCYyHBtDrXdJuXbUzm4A9rKAteGu3Qi5CVR'
        >>> bip38.decrypt(encrypted_wif="6PRVWUbkzzsbcVac2qwfssoUJAN1Xhrg6bNk8J7Nzm5H7kxEbn2Nh2ZoGg", passphrase="TestingOneTwoThree", detail=True)
        {'wif': '5KN7MzqK5wt2TP1fQCYyHBtDrXdJuXbUzm4A9rKAteGu3Qi5CVR', 'private_key': 'cbf4b9f70470856bb4f40f80b87edb90865997ffee6df315ab166d713af433a5', 'wif_type': 'wif', 'public_key': '04d2ce831dd06e5c1f5b1121ef34c2af4bcb01b126e309234adbc3561b60c9360ea7f23327b49ba7f10d17fad15f068b8807dbbc9e4ace5d4a0b40264eefaf31a4', 'public_key_type': 'uncompressed', 'seed': None, 'address': '1Jq6MksXQVWzrznvZzxkV6oY57oWXD9TXB', 'lot': None, 'sequence': None}
        >>> bip38.decrypt(encrypted_wif="6PRL8jj6dLQjBBJjHMdUKLSNLEpjTyAfmt8GnCnfT87NeQ2BU5eAW1tcsS", passphrase="TestingOneTwoThree", network="testnet")
        '938jwjergAxARSWx2YSt9nSBWBz24h8gLhv7EUfgEP1wpMLg6iX'
        >>> bip38.decrypt(encrypted_wif="6PRL8jj6dLQjBBJjHMdUKLSNLEpjTyAfmt8GnCnfT87NeQ2BU5eAW1tcsS", passphrase="TestingOneTwoThree", network="testnet", detail=True)
        {'wif': '938jwjergAxARSWx2YSt9nSBWBz24h8gLhv7EUfgEP1wpMLg6iX', 'private_key': 'cbf4b9f70470856bb4f40f80b87edb90865997ffee6df315ab166d713af433a5', 'wif_type': 'wif', 'public_key': '04d2ce831dd06e5c1f5b1121ef34c2af4bcb01b126e309234adbc3561b60c9360ea7f23327b49ba7f10d17fad15f068b8807dbbc9e4ace5d4a0b40264eefaf31a4', 'public_key_type': 'uncompressed', 'seed': None, 'address': 'myM3eoxWDWxFe7GYHZw8K21rw7QDNZeDYM', 'lot': None, 'sequence': None}
        """

        network: str = (
            network if network else self.network
        )
        if network not in self.networks:
            raise NetworkError("Wrong network type", expected=self.networks, got=type(network))
        encrypted_wif_decode: bytes = decode(encrypted_wif)
        if len(encrypted_wif_decode) != 43:
            raise WIFError("Invalid encrypted WIF length", expected=43, got=len(encrypted_wif_decode))

        prefix: bytes = encrypted_wif_decode[:2]
        flag: bytes = encrypted_wif_decode[2:3]
        address_hash: bytes = encrypted_wif_decode[3:7]

        if prefix == integer_to_bytes(NO_EC_MULTIPLIED_PRIVATE_KEY_PREFIX):

            if flag == integer_to_bytes(NO_EC_MULTIPLIED_WIF_FLAG):
                wif_type: Literal["wif", "wif-compressed"] = "wif"
                public_key_type: str = "uncompressed"
            elif flag == integer_to_bytes(NO_EC_MULTIPLIED_WIF_COMPRESSED_FLAG):
                wif_type: Literal["wif", "wif-compressed"] = "wif-compressed"
                public_key_type: str = "compressed"
            else:
                raise Error(
                    "Invalid flag", expected=[
                        bytes_to_string(integer_to_bytes(NO_EC_MULTIPLIED_WIF_FLAG)),
                        bytes_to_string(integer_to_bytes(NO_EC_MULTIPLIED_WIF_COMPRESSED_FLAG))
                    ], got=bytes_to_string(flag)
                )

            key: bytes = scrypt.hash(
                unicodedata.normalize("NFC", passphrase), address_hash, N=N, r=r, p=p
            )
            derived_half_1, derived_half_2 = key[0:32], key[32:64]
            encrypted_half_1: bytes = encrypted_wif_decode[7:23]
            encrypted_half_2: bytes = encrypted_wif_decode[23:39]

            aes: AESModeOfOperationECB = AESModeOfOperationECB(derived_half_2)
            decrypted_half_2: bytes = aes.decrypt(encrypted_half_2)
            decrypted_half_1: bytes = aes.decrypt(encrypted_half_1)

            private_key: bytes = integer_to_bytes(
                bytes_to_integer(decrypted_half_1 + decrypted_half_2) ^ bytes_to_integer(derived_half_1)
            )
            if bytes_to_integer(private_key) == 0 or bytes_to_integer(private_key) >= NP:
                raise Error("Invalid Non-EC encrypted WIF (Wallet Import Format)")

            public_key: PublicKey = PrivateKey.from_bytes(private_key).public_key()
            address: str = P2PKHAddress.encode(
                public_key=public_key,
                address_prefix=self.cryptocurrency.NETWORKS[network]["address_prefix"],
                public_key_type=public_key_type
            )
            if get_checksum(get_bytes(address, unhexlify=False)) != address_hash:
                raise PassphraseError("Incorrect passphrase")

            wif: str = private_key_to_wif(
                private_key=private_key, wif_type=wif_type, cryptocurrency=self.cryptocurrency, network=network
            )
            if detail:
                return dict(
                    wif=wif,
                    private_key=bytes_to_string(private_key),
                    wif_type=wif_type,
                    public_key=bytes_to_string(public_key.raw(public_key_type=public_key_type)),
                    public_key_type=public_key_type,
                    seed=None,
                    address=address,
                    lot=None,
                    sequence=None
                )
            return wif

        elif prefix == integer_to_bytes(EC_MULTIPLIED_PRIVATE_KEY_PREFIX):
            owner_entropy: bytes = encrypted_wif_decode[7:15]
            encrypted_half_1_half_1: bytes = encrypted_wif_decode[15:23]
            encrypted_half_2: bytes = encrypted_wif_decode[23:-4]

            lot_and_sequence: Optional[bytes] = None
            if bytes_to_integer(flag) in FLAGS["lot_and_sequence"]:
                owner_salt: bytes = owner_entropy[:4]
                lot_and_sequence = owner_entropy[4:]
            else:
                owner_salt: bytes = owner_entropy

            pass_factor: bytes = scrypt.hash(unicodedata.normalize("NFC", passphrase), owner_salt, N=N, r=r, p=p, buflen=32)
            if lot_and_sequence:
                pass_factor: bytes = double_sha256(pass_factor + owner_entropy)
            if bytes_to_integer(pass_factor) == 0 or bytes_to_integer(pass_factor) >= NP:
                raise Error("Invalid EC encrypted WIF (Wallet Import Format)")

            pre_public_key: PublicKey = PrivateKey.from_bytes(pass_factor).public_key()
            salt = address_hash + owner_entropy
            encrypted_seed_b: bytes = scrypt.hash(pre_public_key.raw_compressed(), salt, 1024, 1, 1, 64)
            key: bytes = encrypted_seed_b[32:]

            aes: AESModeOfOperationECB = AESModeOfOperationECB(key)
            encrypted_half_1_half_2_seed_b_last_3 = integer_to_bytes(
                bytes_to_integer(aes.decrypt(encrypted_half_2)) ^ bytes_to_integer(encrypted_seed_b[16:32])
            )
            encrypted_half_1_half_2: bytes = encrypted_half_1_half_2_seed_b_last_3[:8]
            encrypted_half_1: bytes = (
                    encrypted_half_1_half_1 + encrypted_half_1_half_2
            )

            seed_b: bytes = integer_to_bytes(
                bytes_to_integer(aes.decrypt(encrypted_half_1)) ^ bytes_to_integer(encrypted_seed_b[:16])
            ) + encrypted_half_1_half_2_seed_b_last_3[8:]

            factor_b: bytes = double_sha256(seed_b)
            if bytes_to_integer(factor_b) == 0 or bytes_to_integer(factor_b) >= NP:
                raise Error("Invalid EC encrypted WIF (Wallet Import Format)")

            # multiply private key
            private_key: bytes = integer_to_bytes(
                (bytes_to_integer(pass_factor) * bytes_to_integer(factor_b)) % NP
            )
            public_key: PublicKey = PrivateKey.from_bytes(private_key).public_key()
            wif_type: Literal["wif", "wif-compressed"] = "wif"
            public_key_type: str = "uncompressed"
            if bytes_to_integer(flag) in FLAGS["compression"]:
                public_key_type = "compressed"
                wif_type = "wif-compressed"
            address: str = P2PKHAddress.encode(
                public_key=public_key,
                address_prefix=self.cryptocurrency.NETWORKS[network]["address_prefix"],
                public_key_type=public_key_type
            )
            if get_checksum(get_bytes(address, unhexlify=False)) == address_hash:
                wif: str = private_key_to_wif(
                    private_key=private_key, wif_type=wif_type, cryptocurrency=self.cryptocurrency, network=network
                )
                lot: Optional[int] = None
                sequence: Optional[int] = None
                if detail:
                    if lot_and_sequence:
                        sequence: int = bytes_to_integer(lot_and_sequence) % 4096
                        lot: int = (bytes_to_integer(lot_and_sequence) - sequence) // 4096
                    return dict(
                        wif=wif,
                        private_key=bytes_to_string(private_key),
                        wif_type=wif_type,
                        public_key=bytes_to_string(public_key.raw(public_key_type=public_key_type)),
                        public_key_type=public_key_type,
                        seed=bytes_to_string(seed_b),
                        address=address,
                        lot=lot,
                        sequence=sequence
                    )
                return wif
            raise PassphraseError("Incorrect passphrase")
        else:
            raise Error(
                "Invalid prefix", expected=[
                    bytes_to_string(integer_to_bytes(NO_EC_MULTIPLIED_PRIVATE_KEY_PREFIX)),
                    bytes_to_string(integer_to_bytes(EC_MULTIPLIED_PRIVATE_KEY_PREFIX))
                ], got=bytes_to_string(prefix)
            )



#!/usr/bin/env python3

# Copyright © 2023-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import Any
from ecdsa.ecdsa import curve_secp256k1
from ecdsa.ellipticcurve import (
    Point as _Point, PointJacobi
)
from ecdsa import keys

from ..const import COORDINATE_POINT_LENGTH
from ..exceptions import Secp256k1Error
from ..utils import (
    bytes_to_integer, integer_to_bytes
)


class Point:

    point: PointJacobi

    def __init__(self, point_obj: PointJacobi) -> None:
        """
        Initializes the class with a PointJacobi object.

        :param point_obj: The PointJacobi object representing the point.
        :type point_obj: PointJacobi
        """
        self.point = point_obj

    @classmethod
    def from_bytes(cls, point: bytes) -> "Point":
        """
        Creates a point from its byte representation.

        :param point: The byte representation of the point.
        :type point: bytes

        :return: An instance of IPoint representing the decoded point.
        :rtype: IPoint
        """

        try:
            return cls(
                PointJacobi.from_bytes(
                    curve_secp256k1, point
                )
            )
        except keys.MalformedPointError as ex:
            raise Secp256k1Error("Invalid point key bytes") from ex
        except AttributeError:
            return cls.from_coordinates(
                bytes_to_integer(point[:COORDINATE_POINT_LENGTH]),
                bytes_to_integer(point[COORDINATE_POINT_LENGTH:])
            )

    @classmethod
    def from_coordinates(cls, x: int, y: int) -> "Point":
        """
        Creates a point from x and y.

        :param x: The x of the point.
        :type x: int
        :param y: The y of the point.
        :type y: int

        :return: An instance of IPoint representing the point with the given coordinates.
        :rtype: IPoint
        """

        return cls(
            PointJacobi.from_affine(
                _Point(curve_secp256k1, x, y)
            )
        )

    def underlying_object(self) -> Any:
        """
        Returns the underlying point object.

        :return: The underlying point object.
        :rtype: Any
        """

        return self.point

    def x(self) -> int:
        """
        Returns the x of the point.

        :return: The x of the point.
        :rtype: int
        """

        return self.point.x()

    def y(self) -> int:
        """
        Returns the y of the point.

        :return: The y of the point.
        :rtype: int
        """

        return self.point.y()

    def raw(self) -> bytes:
        """
        Returns the raw bytes representation of the point.

        :return: The raw bytes representation of the point.
        :rtype: bytes
        """

        return self.raw_decoded()

    def raw_encoded(self) -> bytes:
        """
        Returns the raw encoded bytes of the point.

        :return: The raw encoded bytes of the point.
        :rtype: bytes
        """

        try:
            return self.point.to_bytes("compressed")
        except AttributeError:
            x: bytes = integer_to_bytes(
                self.point.x(), COORDINATE_POINT_LENGTH
            )
            return b"\x03" + x if self.point.y() & 1 else b"\x02" + x

    def raw_decoded(self) -> bytes:
        """
        Returns the raw decoded bytes of the point.

        :return: The raw bytes of the point.
        :rtype: bytes
        """

        try:
            return self.point.to_bytes()
        except AttributeError:
            x: bytes = integer_to_bytes(
                self.point.x(), COORDINATE_POINT_LENGTH
            )
            y: bytes = integer_to_bytes(
                self.point.y(), COORDINATE_POINT_LENGTH
            )
            return x + y

    def __add__(self, point: "Point") -> "Point":
        """
        Performs addition with another point.

        :param point: The point to add.
        :type point: IPoint

        :return: A new instance of the class representing the result of the addition.
        :rtype: IPoint
        """

        return self.__class__(self.point + point.underlying_object())

    def __radd__(self, point: "Point") -> "Point":
        """
        Performs addition with another point in the context of scalar multiplication.

        :param point: The point to add (right operand).
        :type point: IPoint

        :return: A new instance of the class representing the result of the addition.
        :rtype: IPoint
        """

        return self + point

    def __mul__(self, scalar: int) -> "Point":
        """
        Performs scalar multiplication on the point.

        :param scalar: The scalar value to multiply with the point.
        :type scalar: int

        :return: A new instance of the class representing the result of scalar multiplication.
        :rtype: IPoint
        """

        return self.__class__(self.point * scalar)

    def __rmul__(self, scalar: int) -> "Point":
        """
        Performs scalar multiplication on the point.

        :param scalar: The scalar value to multiply with the point.
        :type scalar: int

        :return: The result of scalar multiplication.
        :rtype: IPoint
        """

        return self * scalar



#!/usr/bin/env python3

# Copyright © 2023-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

import pytest

from bip38.exceptions import Error


def test_exception():
    class TestError(Error):
        pass

    with pytest.raises(TestError, match="Error"):
        raise TestError(message="Error")

    with pytest.raises(TestError, match="more_detail"):
        raise TestError(message="Error", detail="more_detail")

    with pytest.raises(TestError, match="value_expect"):
        raise TestError(message="Error", expected="value_expect")

    with pytest.raises(TestError, match="value_got"):
        raise TestError(message="Error", expected="value_expect", got="value_got")



#!/usr/bin/env python3

# Copyright © 2023-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import Any
from ecdsa import VerifyingKey
from ecdsa.ecdsa import curve_secp256k1
from ecdsa import (
    curves, ellipticcurve, keys
)

from ..const import (
    COMPRESSED_PUBLIC_KEY_LENGTH, UNCOMPRESSED_PUBLIC_KEY_LENGTH
)
from ..exceptions import Secp256k1Error
from .point import Point


class PublicKey:

    verify_key: VerifyingKey

    def __init__(self, verify_key: VerifyingKey) -> None:
        """
        Initialize the instance with a verifying key.

        :param verify_key: The verifying key used for cryptographic operations.
        :type verify_key: VerifyingKey
        """

        self.verify_key = verify_key

    @classmethod
    def from_bytes(cls, public_key: bytes) -> "PublicKey":
        """
        Create a public key instance from bytes representation.

        :param public_key: The bytes representation of the public key.
        :type public_key: bytes

        :return: An instance of the public key derived from the given bytes.
        :rtype: PublicKey
        """

        try:
            return cls(
                VerifyingKey.from_string(
                    public_key, curve=curves.SECP256k1
                )
            )
        except keys.MalformedPointError as ex:
            raise Secp256k1Error("Invalid public key bytes") from ex

    @classmethod
    def from_point(cls, point: Point) -> "PublicKey":
        """
        Create a public key instance from an elliptic curve point.

        :param point: The elliptic curve point representing the public key.
        :type point: Point

        :return: An instance of the public key derived from the given point.
        :rtype: PublicKey
        """

        try:
            return cls(
                VerifyingKey.from_public_point(
                    ellipticcurve.Point(
                        curve_secp256k1, point.x(), point.y()
                    ),
                    curve=curves.SECP256k1
                )
            )
        except keys.MalformedPointError as ex:
            raise Secp256k1Error("Invalid public key point") from ex

    @staticmethod
    def compressed_length() -> int:
        """
        Returns the length of a compressed public key in bytes.

        :return: The length of a compressed public key in bytes.
        :rtype: int
        """

        return COMPRESSED_PUBLIC_KEY_LENGTH

    @staticmethod
    def uncompressed_length() -> int:
        """
        Returns the length of an uncompressed public key in bytes.

        :return: The length of an uncompressed public key in bytes.
        :rtype: int
        """

        return UNCOMPRESSED_PUBLIC_KEY_LENGTH

    def underlying_object(self) -> Any:
        """
        Returns the underlying verification key object.

        :return: The underlying verification key object.
        :rtype: Any
        """

        return self.verify_key

    def raw(self, public_key_type: str) -> bytes:
        """
        Retrieves the raw compressed representation of the public key.

        :return: The compressed public key bytes.
        :rtype: bytes
        """

        return self.raw_compressed() if public_key_type == "compressed" else self.raw_uncompressed()

    def raw_compressed(self) -> bytes:
        """
        Retrieves the raw compressed representation of the public key.

        :return: The compressed public key bytes.
        :rtype: bytes
        """

        return self.verify_key.to_string("compressed")

    def raw_uncompressed(self) -> bytes:
        """
        Retrieves the raw uncompressed representation of the public key.

        :return: The uncompressed public key bytes.
        :rtype: bytes
        """

        return self.verify_key.to_string("uncompressed")

    def point(self) -> Point:
        """
        Retrieves the point object associated with the public key.

        :return: The point object implementing the Point interface.
        :rtype: Point
        """

        return Point(self.verify_key.pubkey.point)



#!/usr/bin/env python3

import json

from bip38 import BIP38
from bip38.cryptocurrencies import Bitcoin as Cryptocurrency

# Define the passphrase used for encryption/decryption
PASSPHRASE: str = "meherett"
# Specify the Wallet Import Format (WIF)
WIF: str = "L44B5gGEpqEDRS9vVPz7QT35jcBG2r3CZwSwQ4fCewXAhAhqGVpP"  # wif-compressed type
# Define the network type
NETWORK: str = "mainnet"
# Whether to show detailed information during operations
DETAIL: bool = True

# Initialize the BIP38 object with the cryptocurrency and network
bip38: BIP38 = BIP38(
    cryptocurrency=Cryptocurrency, network=NETWORK
)
# Encrypt the WIF using the BIP38 standard and the specified passphrase
encrypted_wif: str = bip38.encrypt(
    wif=WIF, passphrase=PASSPHRASE
)
print("BIP38 Encrypted WIF:", encrypted_wif)
# Decrypt the BIP38 encrypted WIF back to the original private key and show details
print("BIP38 Decrypted:", json.dumps(bip38.decrypt(
    encrypted_wif=encrypted_wif, passphrase=PASSPHRASE, detail=DETAIL
), indent=4))



#!/usr/bin/env python3

# Copyright © 2023-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    AnyStr, Optional, Union, Literal
)


def get_bytes(data: AnyStr, unhexlify: bool = True) -> bytes:
    """
    Convert input data to bytes format.

    :param data: The input data to convert. Can be bytes or string.
    :type data: Union[bytes, str]
    :param unhexlify: Flag indicating whether to interpret strings as hexadecimal (default True).
    :type unhexlify: bool

    :return: The input data converted to bytes format.
    :rtype: bytes
    """

    if not data:
        return b''
    if isinstance(data, bytes):
        return data
    elif isinstance(data, str):
        if unhexlify:
            return bytes.fromhex(data)
        else:
            return bytes(data, 'utf-8')
    else:
        raise TypeError("Agreement must be either 'bytes' or 'string'!")


def bytes_reverse(data: bytes) -> bytes:
    """
    Reverse the order of bytes in a bytes object.

    :param data: The bytes object to reverse.
    :type data: bytes

    :return: The bytes object with its byte order reversed.
    :rtype: bytes
    """

    tmp = bytearray(data)
    tmp.reverse()
    return bytes(tmp)


def bytes_to_string(data: Union[bytes, str]) -> str:
    """
    Convert bytes or string data to a hexadecimal string representation.

    :param data: The bytes or string data to convert to hexadecimal string.
    :type data: Union[bytes, str]

    :return: The hexadecimal string representation of the input data.
    :rtype: str
    """

    if not data:
        return ''
    try:
        bytes.fromhex(data)
        return data
    except (ValueError, TypeError):
        pass
    if not isinstance(data, bytes):
        data = bytes(data, 'utf-8')
    return data.hex()


def bytes_to_integer(data: bytes, endianness: Literal["little", "big"] = "big", signed: bool = False) -> int:
    """
    Convert bytes to an integer based on specified endianness and signedness.

    :param data: The bytes object to convert to an integer.
    :type data: bytes
    :param endianness: The byte order ("little" or "big").
    :type endianness: Literal["little", "big"]
    :param signed: Flag indicating whether the integer is signed (default False).
    :type signed: bool

    :return: The integer value converted from bytes.
    :rtype: int
    """

    return int.from_bytes(data, byteorder=endianness, signed=signed)


def integer_to_bytes(
    data: int, bytes_num: Optional[int] = None, endianness: Literal["little", "big"] = "big", signed: bool = False
) -> bytes:
    """
    Convert an integer to bytes based on specified parameters.

    :param data: The integer to convert to bytes.
    :type data: int
    :param bytes_num: Optional number of bytes to use for the conversion. If not provided, it is calculated based on the integer's bit length.
    :type bytes_num: Optional[int]
    :param endianness: The byte order ("little" or "big").
    :type endianness: Literal["little", "big"]
    :param signed: Flag indicating whether the integer is signed (default False).
    :type signed: bool

    :return: The bytes object representing the integer.
    :rtype: bytes
    """

    bytes_num = bytes_num or ((data.bit_length() if data > 0 else 1) + 7) // 8
    return data.to_bytes(bytes_num, byteorder=endianness, signed=signed)



#!/usr/bin/env python3

# Copyright © 2023-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    List, Dict
)

# non-EC-multiplied & EC-multiplied private key prefixes
NO_EC_MULTIPLIED_PRIVATE_KEY_PREFIX: int = 0x0142
EC_MULTIPLIED_PRIVATE_KEY_PREFIX: int = 0x0143
# non-EC-multiplied Wallet Import Format (WIF) flags
NO_EC_MULTIPLIED_WIF_FLAG: int = 0xc0
NO_EC_MULTIPLIED_WIF_COMPRESSED_FLAG: int = 0xe0
# Magic bytes for lot and sequence and non lot and sequence
MAGIC_LOT_AND_SEQUENCE: int = 0x2ce9b3e1ff39e251
MAGIC_NO_LOT_AND_SEQUENCE: int = 0x2ce9b3e1ff39e253
# Magic uncompressed and compressed flags
MAGIC_LOT_AND_SEQUENCE_UNCOMPRESSED_FLAG: int = 0x04
MAGIC_LOT_AND_SEQUENCE_COMPRESSED_FLAG: int = 0x24
MAGIC_NO_LOT_AND_SEQUENCE_UNCOMPRESSED_FLAG: int = 0x00
MAGIC_NO_LOT_AND_SEQUENCE_COMPRESSED_FLAG: int = 0x20
# Confirmation code prefix
CONFIRMATION_CODE_PREFIX: int = 0x643bf6a89a
# Number of points in the field
NP: int = 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFEBAAEDCE6AF48A03BBFD25E8CD0364141
# Coordinate point length
COORDINATE_POINT_LENGTH: int = 32
# Private key length & prefixes
PRIVATE_KEY_LENGTH: int = 32
UNCOMPRESSED_PRIVATE_KEY_PREFIX: int = 0x00
COMPRESSED_PRIVATE_KEY_PREFIX: int = 0x01
# Public key length & prefixes
COMPRESSED_PUBLIC_KEY_LENGTH: int = 33
UNCOMPRESSED_PUBLIC_KEY_LENGTH: int = 65
EVEN_COMPRESSED_PUBLIC_KEY_PREFIX: int = 0x02
ODD_COMPRESSED_PUBLIC_KEY_PREFIX: int = 0x03
UNCOMPRESSED_PUBLIC_KEY_PREFIX: int = 0x04
# Checksum byte length
CHECKSUM_BYTE_LENGTH: int = 4
# Wallet Import Format (WIF) types
WIF_TYPES = ["wif", "wif-compressed"]
# Public Key types
PUBLIC_KEY_TYPES = ["uncompressed", "compressed"]
# List of compression, lot_and_sequence, non_ec, ec, & illegal flags
FLAGS: Dict[str, List[int]] = {
    "compression": [
        MAGIC_NO_LOT_AND_SEQUENCE_COMPRESSED_FLAG, MAGIC_LOT_AND_SEQUENCE_COMPRESSED_FLAG, 0x28, 0x2c, 0x30, 0x34, 0x38, 0x3c, 0xe0, 0xe8, 0xf0, 0xf8
    ],
    "lot_and_sequence": [
        MAGIC_LOT_AND_SEQUENCE_UNCOMPRESSED_FLAG, MAGIC_LOT_AND_SEQUENCE_COMPRESSED_FLAG, 0x0c, 0x14, 0x1c, 0x2c, 0x34, 0x3c
    ],
    "non_ec": [
        NO_EC_MULTIPLIED_WIF_FLAG, NO_EC_MULTIPLIED_WIF_COMPRESSED_FLAG, 0xc8, 0xd0, 0xd8, 0xe8, 0xf0, 0xf8
    ],
    "ec": [
        0x00, 0x04, 0x08, 0x0c, 0x10, 0x14, 0x18, 0x1c, 0x20, 0x24, 0x28, 0x2c, 0x30, 0x34, 0x38, 0x3c
    ],
    "illegal": [
        0xc4, 0xcc, 0xd4, 0xdc, 0xe4, 0xec, 0xf4, 0xfc
    ]
}



#!/usr/bin/env python3

# Copyright © 2023-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from bip38.p2pkh_address import P2PKHAddress


def test_p2pkh_address(_):

    assert P2PKHAddress.encode(
        public_key=_["p2pkh_address"]["compressed"]["public_key"],
        public_key_address_prefix=int(_["p2pkh_address"]["compressed"]["args"]["public_key_address_prefix"], base=16),
        public_key_type=_["p2pkh_address"]["compressed"]["args"]["public_key_type"]
    ) == _["p2pkh_address"]["compressed"]["encode"]

    assert P2PKHAddress.decode(
        address=_["p2pkh_address"]["compressed"]["encode"],
        public_key_address_prefix=int(_["p2pkh_address"]["compressed"]["args"]["public_key_address_prefix"], base=16),
        public_key_type=_["p2pkh_address"]["compressed"]["args"]["public_key_type"]
    ) == _["p2pkh_address"]["compressed"]["decode"]

    assert P2PKHAddress.encode(
        public_key=_["p2pkh_address"]["uncompressed"]["public_key"],
        public_key_address_prefix=int(_["p2pkh_address"]["uncompressed"]["args"]["public_key_address_prefix"], base=16),
        public_key_type=_["p2pkh_address"]["uncompressed"]["args"]["public_key_type"]
    ) == _["p2pkh_address"]["uncompressed"]["encode"]

    assert P2PKHAddress.decode(
        address=_["p2pkh_address"]["uncompressed"]["encode"],
        public_key_address_prefix=int(_["p2pkh_address"]["uncompressed"]["args"]["public_key_address_prefix"], base=16),
        public_key_type=_["p2pkh_address"]["uncompressed"]["args"]["public_key_type"]
    ) == _["p2pkh_address"]["uncompressed"]["decode"]



#!/usr/bin/env python3

# Copyright © 2023-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

import os

from bip38.utils import (
    get_bytes, bytes_reverse, bytes_to_string, bytes_to_integer, integer_to_bytes
)


def test_utils(_):

    assert isinstance(get_bytes(data=os.urandom(8), unhexlify=False), bytes)
    assert isinstance(get_bytes(data=os.urandom(8), unhexlify=False), bytes)
    assert isinstance(get_bytes(data=bytes_to_string(os.urandom(8)), unhexlify=True), bytes)
    assert isinstance(get_bytes(data=bytes_to_string(os.urandom(8)), unhexlify=True), bytes)
    assert isinstance(bytes_reverse(data=os.urandom(8)), bytes)
    assert isinstance(bytes_to_string(data=str()), str)
    assert isinstance(bytes_to_string(data=os.urandom(8)), str)
    assert isinstance(bytes_to_integer(data=os.urandom(8)), int)
    assert isinstance(integer_to_bytes(bytes_to_integer(data=os.urandom(8))), bytes)



#!/usr/bin/env python3

# Copyright © 2023-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import List

__name__: str = "bip38"
__version__: str = "v1.4.0"
__license__: str = "MIT"
__author__: str = "Meheret Tesfaye Batu"
__email__: str = "meherett.batu@gmail.com"
__documentation__: str = "https://bip38.readthedocs.com"
__description__: str = "A Python library for implementation of Bitcoin Improvement Proposal - 0038 / BIP38 protocol."
__url__: str = "https://hdwallet.io"
__source__: str = "https://github.com/hdwallet-io/python-bip38"
__changelog__: str = f"{__source__}/blob/master/CHANGELOG.md"
__tracker__: str = f"{__url__}/issues"
__keywords__: List[str] = [
    "bip38", "bitcoin", "private-key", "hdwallet", "encrypt", "decrypt", "passphrase", "wif", "security"
]
__websites__: List[str] = [
    "https://talonlab.org",
    "https://talonlab.gitbook.io/bip38",
    __documentation__,
    "https://hdwallet.online",
    "https://hd.wallet",  # On Web3 domain
    __url__
]


#!/usr/bin/env python3

# Copyright © 2023-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit


from bip38.crypto import (
    ripemd160, sha256, double_sha256, get_checksum, hash160
)
from bip38.utils import get_bytes


def test_crypto(_):

    assert isinstance(ripemd160(data=_["crypto"]["data"]), bytes)
    assert ripemd160(data=_["crypto"]["data"]) == get_bytes(_["crypto"]["ripemd160"])

    assert isinstance(sha256(data=_["crypto"]["data"]), bytes)
    assert sha256(data=_["crypto"]["data"]) == get_bytes(_["crypto"]["sha256"])

    assert isinstance(double_sha256(data=_["crypto"]["data"]), bytes)
    assert double_sha256(data=_["crypto"]["data"]) == get_bytes(_["crypto"]["double_sha256"])

    assert isinstance(get_checksum(data=_["crypto"]["data"]), bytes)
    assert get_checksum(data=_["crypto"]["data"]) == get_bytes(_["crypto"]["get_checksum"])

    assert isinstance(hash160(data=_["crypto"]["data"]), bytes)
    assert hash160(data=_["crypto"]["data"]) == get_bytes(_["crypto"]["hash160"])



#!/usr/bin/env python3

# Copyright © 2023-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from ecdsa.ecdsa import generator_secp256k1

from .point import Point
from .public_key import PublicKey
from .private_key import PrivateKey


class Secp256k1:  # Elliptic Curve Cryptography (ECC)

    ORDER: str = generator_secp256k1.order()
    GENERATOR = Point(generator_secp256k1)
    POINT = Point
    PUBLIC_KEY = PublicKey
    PRIVATE_KEY = PrivateKey



#!/usr/bin/env python3

from typing import (
    Union, Literal, Optional
)

import json
import os

from bip38 import BIP38
from bip38.cryptocurrencies import Bitcoin as Cryptocurrency


# Define the passphrase used for encryption/decryption
PASSPHRASE: str = "meherett"
# Pick a random owner salt or use your own
OWNER_SALT: Union[str, bytes] = os.urandom(8)  # Example: "75ed1cdeb254cb38"
# Pick a random seed or use your own
SEED: Union[str, bytes] = os.urandom(24)  # Example: "99241d58245c883896f80843d2846672d7312e6195ca1a6c"
# Specify the WIF type
WIF_TYPE: Literal["wif", "wif-compressed"] = "wif-compressed"
# Define the network type
NETWORK: str = "mainnet"
# Specify the lot number (100000 <= lot <= 999999), or set to None
LOT: Optional[int] = None
# Specify the sequence number (0 <= sequence <= 4095), or set to None
SEQUENCE: Optional[int] = None
# Whether to show detailed information during operations
DETAIL: bool = True

# Initialize the BIP38 object with the cryptocurrency and network
bip38: BIP38 = BIP38(
    cryptocurrency=Cryptocurrency, network=NETWORK
)
# Generate the intermediate passphrase (used in BIP38 encryption)
intermediate_passphrase: str = bip38.intermediate_code(
    passphrase=PASSPHRASE, owner_salt=OWNER_SALT, lot=LOT, sequence=SEQUENCE
)
print("Intermediate Passphrase:", intermediate_passphrase)
# Create a new encrypted WIF (Wallet Import Format) using the intermediate passphrase
encrypted_wif: dict = bip38.create_new_encrypted_wif(
    intermediate_passphrase=intermediate_passphrase, wif_type=WIF_TYPE, seed=SEED
)
print("Encrypted WIF:", json.dumps(encrypted_wif, indent=4))
# Confirm the encrypted WIF using the passphrase and confirmation code
print("Confirm Code:", json.dumps(bip38.confirm_code(
    passphrase=PASSPHRASE, confirmation_code=encrypted_wif["confirmation_code"], detail=DETAIL
), indent=4))
# Decrypt the BIP38 encrypted WIF back to the original private key
print("BIP38 Decrypted:", json.dumps(bip38.decrypt(
    encrypted_wif=encrypted_wif["encrypted_wif"], passphrase=PASSPHRASE, detail=DETAIL
), indent=4))



#!/usr/bin/env python3

# Copyright © 2023-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

import pytest

from bip38.bip38 import BIP38
from bip38.cryptocurrencies import Bitcoin
from bip38.exceptions import (
    Error, NetworkError, PassphraseError
)


def test_intermediate_code(_):

    for index in range(len(_["bip38"]["intermediate_code"])):
        bip38: BIP38 = BIP38(
            cryptocurrency=Bitcoin, network=_["bip38"]["decrypt"][index]["network"]
        )
        intermediate_passphrase: str = bip38.intermediate_code(
            passphrase=_["bip38"]["intermediate_code"][index]["passphrase"],
            lot=_["bip38"]["intermediate_code"][index]["lot"],
            sequence=_["bip38"]["intermediate_code"][index]["sequence"],
            owner_salt=_["bip38"]["intermediate_code"][index]["owner_salt"],
        )

        assert isinstance(intermediate_passphrase, str)

        assert intermediate_passphrase == _["bip38"]["intermediate_code"][index]["intermediate_passphrase"]


def test_bip38_encrypt(_):

    for index in range(len(_["bip38"]["encrypt"])):
        bip38: BIP38 = BIP38(
            cryptocurrency=Bitcoin, network=_["bip38"]["encrypt"][index]["network"]
        )
        encrypted_wif: str = bip38.encrypt(
            wif=_["bip38"]["encrypt"][index]["wif"],
            passphrase=_["bip38"]["encrypt"][index]["passphrase"]
        )

        assert isinstance(encrypted_wif, str)
        assert encrypted_wif == _["bip38"]["encrypt"][index]["encrypted_wif"]

        with pytest.raises(NetworkError):
            bip38.encrypt(
                wif=_["bip38"]["encrypt"][index]["wif"],
                passphrase=_["bip38"]["encrypt"][index]["passphrase"],
                network="FAKE_NETWORK"
            )


def test_create_new_encrypted_wif(_):

    for index in range(len(_["bip38"]["create_new_encrypted_wif"])):
        bip38: BIP38 = BIP38(
            cryptocurrency=Bitcoin, network="mainnet"
        )
        encrypted_wif: dict = bip38.create_new_encrypted_wif(
            intermediate_passphrase=_["bip38"]["create_new_encrypted_wif"][index]["intermediate_passphrase"],
            wif_type=_["bip38"]["create_new_encrypted_wif"][index]["wif_type"],
            seed=_["bip38"]["create_new_encrypted_wif"][index]["seed"]
        )

        assert isinstance(encrypted_wif, dict)

        assert encrypted_wif["encrypted_wif"] == _["bip38"]["create_new_encrypted_wif"][index]["encrypted_wif"]
        assert encrypted_wif["confirmation_code"] == _["bip38"]["create_new_encrypted_wif"][index]["confirmation_code"]
        assert encrypted_wif["public_key"] == _["bip38"]["create_new_encrypted_wif"][index]["public_key"]
        assert encrypted_wif["public_key_type"] == _["bip38"]["create_new_encrypted_wif"][index]["public_key_type"]
        assert encrypted_wif["address"] == _["bip38"]["create_new_encrypted_wif"][index]["address"]

        with pytest.raises(PassphraseError):
            bip38.create_new_encrypted_wif(
                intermediate_passphrase="FAKE_PASSPHRASE",
                wif_type=_["bip38"]["create_new_encrypted_wif"][index]["wif_type"],
                seed=_["bip38"]["create_new_encrypted_wif"][index]["seed"]
            )
            
        with pytest.raises(NetworkError):
            bip38.create_new_encrypted_wif(
                intermediate_passphrase=_["bip38"]["create_new_encrypted_wif"][index]["intermediate_passphrase"],
                wif_type=_["bip38"]["create_new_encrypted_wif"][index]["wif_type"],
                seed=_["bip38"]["create_new_encrypted_wif"][index]["seed"],
                network="FAKE_NETWORK"
            )


def test_confirm_code(_):

    for index in range(len(_["bip38"]["confirm_code"])):
        bip38: BIP38 = BIP38(
            cryptocurrency=Bitcoin, network="mainnet"
        )
        confirmed: str = bip38.confirm_code(
            passphrase=_["bip38"]["confirm_code"][index]["passphrase"],
            confirmation_code=_["bip38"]["confirm_code"][index]["confirmation_code"],
            detail=False
        )

        assert isinstance(confirmed, str)

        assert confirmed == _["bip38"]["confirm_code"][index]["address"]

        confirmed: dict = bip38.confirm_code(
            passphrase=_["bip38"]["confirm_code"][index]["passphrase"],
            confirmation_code=_["bip38"]["confirm_code"][index]["confirmation_code"],
            detail=True
        )

        assert isinstance(confirmed, dict)

        assert confirmed["public_key"] == _["bip38"]["confirm_code"][index]["public_key"]
        assert confirmed["public_key_type"] == _["bip38"]["confirm_code"][index]["public_key_type"]
        assert confirmed["address"] == _["bip38"]["confirm_code"][index]["address"]
        assert confirmed["lot"] == _["bip38"]["confirm_code"][index]["lot"]
        assert confirmed["sequence"] == _["bip38"]["confirm_code"][index]["sequence"]

        with pytest.raises(NetworkError):
            bip38.confirm_code(
                passphrase=_["bip38"]["confirm_code"][index]["passphrase"],
                confirmation_code=_["bip38"]["confirm_code"][index]["confirmation_code"],
                detail=False,
                network="FAKE_NETWORK"
            )

        with pytest.raises(Error):
            bip38.confirm_code(
                passphrase=_["bip38"]["confirm_code"][index]["passphrase"],
                confirmation_code="FAKE_CONFIRMATION_CODE",
                detail=False
            )

        with pytest.raises(PassphraseError):
            bip38.confirm_code(
                passphrase="FAKE_PASSPHRASE",
                confirmation_code=_["bip38"]["confirm_code"][index]["confirmation_code"],
                detail=False
            )


def test_bip38_decrypt(_):

    for index in range(len(_["bip38"]["decrypt"])):
        bip38: BIP38 = BIP38(
            cryptocurrency=Bitcoin, network=_["bip38"]["decrypt"][index]["network"]
        )
        decrypted_wif: str = bip38.decrypt(
            encrypted_wif=_["bip38"]["decrypt"][index]["encrypted_wif"],
            passphrase=_["bip38"]["decrypt"][index]["passphrase"],
            detail=False
        )

        assert isinstance(decrypted_wif, str)

        assert decrypted_wif == _["bip38"]["decrypt"][index]["wif"]

        decrypted: dict = bip38.decrypt(
            encrypted_wif=_["bip38"]["decrypt"][index]["encrypted_wif"],
            passphrase=_["bip38"]["decrypt"][index]["passphrase"],
            detail=True
        )

        assert isinstance(decrypted, dict)

        assert decrypted["wif"] == _["bip38"]["decrypt"][index]["wif"]
        assert decrypted["private_key"] == _["bip38"]["decrypt"][index]["private_key"]
        assert decrypted["wif_type"] == _["bip38"]["decrypt"][index]["wif_type"]
        assert decrypted["public_key"] == _["bip38"]["decrypt"][index]["public_key"]
        assert decrypted["public_key_type"] == _["bip38"]["decrypt"][index]["public_key_type"]
        assert decrypted["seed"] == _["bip38"]["decrypt"][index]["seed"]
        assert decrypted["address"] == _["bip38"]["decrypt"][index]["address"]
        assert decrypted["lot"] == _["bip38"]["decrypt"][index]["lot"]
        assert decrypted["sequence"] == _["bip38"]["decrypt"][index]["sequence"]

        with pytest.raises(NetworkError):
            bip38.decrypt(
                encrypted_wif=_["bip38"]["decrypt"][index]["encrypted_wif"],
                passphrase=_["bip38"]["decrypt"][index]["passphrase"],
                detail=True,
                network="FAKE_NETWORK"
            )



#!/usr/bin/env python3

# Copyright © 2020-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    Optional, Any
)


class Error(Exception):

    def __init__(
        self, message: str, detail: Optional[str] = None, expected: Any = None, got: Any = None
    ) -> None:
        self._message, self._detail, self._expected, self._got = (
            message, detail, None, f"'{got}'"
        )

        if isinstance(expected, list):
            for expect in expected:
                if self._expected is None:
                    self._expected = f"'{expect}'"
                else:
                    self._expected += f", '{expect}'"
        else:
            self._expected = expected

    def __str__(self):
        if self._expected and self._got and self._detail:
            return f"{self._message}, (expected: {self._expected} | got: {self._got}) {self._detail}"
        elif self._expected and self._got and not self._detail:
            return f"{self._message}, (expected: {self._expected} | got: {self._got})"
        elif self._detail:
            return f"{self._message} {self._detail}"
        else:
            return f"{self._message}"


class CryptocurrencyError(Error):
    pass


class NetworkError(Error):
    pass


class Secp256k1Error(Error):
    pass


class AddressError(Error):
    pass


class WIFError(Error):
    pass


class PassphraseError(Error):
    pass



#!/usr/bin/env python3

# Copyright © 2023-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import Union

import hashlib

from .libs.ripemd160 import ripemd160 as r160
from .const import CHECKSUM_BYTE_LENGTH
from .utils import get_bytes


def ripemd160(data: Union[str, bytes]) -> bytes:
    """
    Calculate the RIPEMD-160 hash of the given data.

    :param data: The data to hash, as bytes or a string.
    :type data: Union[str, bytes]

    :return: The RIPEMD-160 hash digest as bytes.
    :rtype: bytes
    """

    return (
        hashlib.new("ripemd160", get_bytes(data)).digest()
        if "ripemd160" in hashlib.algorithms_available else
        r160(get_bytes(data))
    )


def sha256(data: Union[str, bytes]) -> bytes:
    """
    Calculate the SHA-256 hash of the given data.

    :param data: The data to hash, as bytes or a string.
    :type data: Union[str, bytes]

    :return: The SHA-256 hash digest as bytes.
    :rtype: bytes
    """

    return hashlib.sha256(get_bytes(data)).digest()


def double_sha256(data: Union[str, bytes]) -> bytes:
    """
    Calculate the double SHA-256 hash of the given data.

    :param data: The data to hash, as bytes or a string.
    :type data: Union[str, bytes]

    :return: The double SHA-256 hash digest as bytes.
    :rtype: bytes
    """

    return sha256(sha256(data))


def get_checksum(data: Union[str, bytes]) -> bytes:
    """
    Calculate the checksum for the given raw bytes.

    The checksum is derived by performing a double SHA-256 hash on the input
    and returning the first few bytes, as determined by `CHECKSUM_BYTE_LENGTH`.

    :param data: The raw data to checksum.
    :type data: Union[str, bytes]

    :returns: The checksum of the data.
    :rtype: bytes
    """

    return double_sha256(data)[:CHECKSUM_BYTE_LENGTH]


def hash160(data: Union[str, bytes]) -> bytes:
    """
    Calculate the HASH160 hash (RIPEMD-160 of SHA-256) of the given data.

    :param data: The data to hash, as bytes or a string.
    :type data: Union[str, bytes]

    :return: The HASH160 hash digest as bytes.
    :rtype: bytes
    """

    return ripemd160(sha256(data))



#!/usr/bin/env python3

# Copyright © 2023-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from ecdsa import (
    SigningKey, VerifyingKey
)
from ecdsa.ellipticcurve import PointJacobi

from bip38.secp256k1 import (
    Secp256k1, Point, PublicKey, PrivateKey
)
from bip38.utils import get_bytes


def test_secp256k1():

    assert isinstance(Secp256k1.ORDER, int)
    assert isinstance(Secp256k1.GENERATOR, Point)
    assert isinstance(Secp256k1.POINT, type(Point))
    assert isinstance(Secp256k1.PUBLIC_KEY, type(PublicKey))
    assert isinstance(Secp256k1.PRIVATE_KEY, type(PrivateKey))


def test_secp256k1_point(_):

    for public_key_type in ["uncompressed", "compressed"]:
        # Test from bytes
        point = Point.from_bytes(
            get_bytes(_["secp256k1"][public_key_type]["point"]["encode"])
        )

        assert isinstance(point, Point)
        assert isinstance(point.underlying_object(), PointJacobi)

        assert point.x() == _["secp256k1"][public_key_type]["point"]["x"]
        assert point.y() == _["secp256k1"][public_key_type]["point"]["y"]
        assert point.raw_encoded() == get_bytes(_["secp256k1"][public_key_type]["point"]["encode"])
        assert point.raw() == point.raw_decoded() == get_bytes(_["secp256k1"][public_key_type]["point"]["decode"])

        # Test from coordinate
        point = Point.from_coordinates(
            x=_["secp256k1"][public_key_type]["point"]["x"],
            y=_["secp256k1"][public_key_type]["point"]["y"]
        )

        assert point.x() == _["secp256k1"][public_key_type]["point"]["x"]
        assert point.y() == _["secp256k1"][public_key_type]["point"]["y"]
        assert point.raw_encoded() == get_bytes(_["secp256k1"][public_key_type]["point"]["encode"])
        assert point.raw() == point.raw_decoded() == get_bytes(_["secp256k1"][public_key_type]["point"]["decode"])


def test_secp256k1_public_key(_):

    assert PublicKey.uncompressed_length() == _["secp256k1"]["uncompressed"]["length"]
    assert PublicKey.compressed_length() == _["secp256k1"]["compressed"]["length"]
    for public_key_type in ["uncompressed", "compressed"]:
        public_key = PublicKey.from_bytes(
            get_bytes(_["secp256k1"][public_key_type]["public_key"])
        )
        assert isinstance(public_key, PublicKey)
        assert isinstance(public_key.point(), Point)
        assert isinstance(public_key.underlying_object(), VerifyingKey)

        assert public_key.raw_uncompressed() == get_bytes(_["secp256k1"]["uncompressed"]["public_key"])
        assert public_key.raw_compressed() == get_bytes(_["secp256k1"]["compressed"]["public_key"])


def test_secp256k1_private_key(_):

    assert PrivateKey.length() == _["secp256k1"]["private_key_length"]
    private_key = PrivateKey.from_bytes(
        get_bytes(_["secp256k1"]["private_key"])
    )

    assert isinstance(private_key, PrivateKey)
    assert isinstance(private_key.underlying_object(), SigningKey)
    assert isinstance(private_key.raw(), bytes)
    assert isinstance(private_key.public_key(), PublicKey)

    assert private_key.raw() == get_bytes(_["secp256k1"]["private_key"])
    assert private_key.public_key().raw_uncompressed() == get_bytes(_["secp256k1"]["uncompressed"]["public_key"])
    assert private_key.public_key().raw_compressed() == get_bytes(_["secp256k1"]["compressed"]["public_key"])



#!/usr/bin/env python3

# Copyright © 2023-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    Dict, Optional
)


class ICryptocurrency:

    NETWORKS: Dict[str, Dict[str, int]]
    ALPHABET: Optional[str] = None


class Adcoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xb0,
            "address_prefix": 0x17
        }
    }


class Anon(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x80,
            "address_prefix": 0x582
        }
    }


class Argoneum(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xbf,
            "address_prefix": 0x32
        }
    }


class Artax(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x97,
            "address_prefix": 0x17
        }
    }


class Aryacoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x97,
            "address_prefix": 0x17
        }
    }


class Asiacoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x97,
            "address_prefix": 0x17
        }
    }


class Auroracoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x97,
            "address_prefix": 0x17
        }
    }


class Avian(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x80,
            "address_prefix": 0x3c
        }
    }


class Axe(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xcc,
            "address_prefix": 0x37
        }
    }


class Bata(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xa4,
            "address_prefix": 0x19
        }
    }


class BeetleCoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x99,
            "address_prefix": 0x1a
        }
    }


class BelaCoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x99,
            "address_prefix": 0x19
        }
    }


class BitCloud(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x99,
            "address_prefix": 0x19
        }
    }


class Bitcoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x80,
            "address_prefix": 0x0
        },
        "testnet": {
            "wif_prefix": 0xef,
            "address_prefix": 0x6f
        },
        "regtest": {
            "wif_prefix": 0xef,
            "address_prefix": 0x6f
        }
    }
    ALPHABET = "123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz"


class BitcoinAtom(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x80,
            "address_prefix": 0x17
        }
    }


class BitcoinGold(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x80,
            "address_prefix": 0x26
        }
    }


class BitcoinGreen(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x2e,
            "address_prefix": 0x26
        }
    }


class BitcoinPlus(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x99,
            "address_prefix": 0x19
        }
    }


class BitcoinPrivate(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x80,
            "address_prefix": 0x1325
        },
        "testnet": {
            "wif_prefix": 0xef,
            "address_prefix": 0x1957
        }
    }


class BitcoinSV(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x80,
            "address_prefix": 0x0
        }
    }


class BitcoinZ(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x80,
            "address_prefix": 0x1cb8
        }
    }


class Bitcore(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x80,
            "address_prefix": 0x3
        }
    }


class BitSend(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xcc,
            "address_prefix": 0x66
        }
    }


class Blackcoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x99,
            "address_prefix": 0x19
        }
    }


class Blocknode(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x4b,
            "address_prefix": 0x19
        },
        "testnet": {
            "wif_prefix": 0x89,
            "address_prefix": 0x55
        }
    }


class BlockStamp(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x80,
            "address_prefix": 0x0
        }
    }


class Bolivarcoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xd5,
            "address_prefix": 0x55
        }
    }


class BritCoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x99,
            "address_prefix": 0x19
        }
    }


class CanadaeCoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x9c,
            "address_prefix": 0x1c
        }
    }


class Cannacoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x9c,
            "address_prefix": 0x1c
        }
    }


class Clams(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x85,
            "address_prefix": 0x89
        }
    }


class ClubCoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x99,
            "address_prefix": 0x1c
        }
    }


class Compcoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x9c,
            "address_prefix": 0x1c
        }
    }


class CPUChain(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x80,
            "address_prefix": 0x1c
        }
    }


class CranePay(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x7b,
            "address_prefix": 0x1c
        }
    }


class Crave(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x99,
            "address_prefix": 0x46
        }
    }


class Dash(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xcc,
            "address_prefix": 0x4c
        },
        "testnet": {
            "wif_prefix": 0xef,
            "address_prefix": 0x8c
        }
    }


class DeepOnion(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x9f,
            "address_prefix": 0x1f
        }
    }


class Defcoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x9e,
            "address_prefix": 0x1e
        }
    }


class Denarius(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x9e,
            "address_prefix": 0x1e
        }
    }


class Diamond(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xda,
            "address_prefix": 0x5a
        }
    }


class DigiByte(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x80,
            "address_prefix": 0x1e
        }
    }


class Digitalcoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x9e,
            "address_prefix": 0x1e
        }
    }


class Divi(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xd4,
            "address_prefix": 0x1e
        },
        "testnet": {
            "wif_prefix": 0xd4,
            "address_prefix": 0x1e
        }
    }


class Dogecoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xf1,
            "address_prefix": 0x1e
        },
        "testnet": {
            "wif_prefix": 0xf1,
            "address_prefix": 0x71
        }
    }


class Ecoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xdc,
            "address_prefix": 0x5c
        }
    }


class EDRCoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xdd,
            "address_prefix": 0x5d
        }
    }


class eGulden(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xb0,
            "address_prefix": 0x30
        }
    }


class Einsteinium(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xa1,
            "address_prefix": 0x21
        }
    }


class Elastos(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x80,
            "address_prefix": 0x21
        }
    }


class Energi(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x6a,
            "address_prefix": 0x21
        }
    }


class EuropeCoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xa8,
            "address_prefix": 0x21
        }
    }


class Evrmore(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x80,
            "address_prefix": 0x21
        },
        "testnet": {
            "wif_prefix": 0xef,
            "address_prefix": 0x6f
        }
    }


class ExclusiveCoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xa1,
            "address_prefix": 0x21
        }
    }


class Feathercoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x8e,
            "address_prefix": 0xe
        }
    }


class Firo(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xd2,
            "address_prefix": 0x52
        }
    }


class Firstcoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xa3,
            "address_prefix": 0x23
        }
    }


class FIX(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x3c,
            "address_prefix": 0x23
        },
        "testnet": {
            "wif_prefix": 0xed,
            "address_prefix": 0x4c
        }
    }


class Flashcoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xc4,
            "address_prefix": 0x44
        }
    }


class Flux(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x80,
            "address_prefix": 0x1cb8
        }
    }


class Foxdcoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x80,
            "address_prefix": 0x23
        },
        "testnet": {
            "wif_prefix": 0xef,
            "address_prefix": 0x5f
        }
    }


class FujiCoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xa4,
            "address_prefix": 0x24
        }
    }


class GameCredits(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xa6,
            "address_prefix": 0x26
        }
    }


class GCRCoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x9a,
            "address_prefix": 0x26
        }
    }


class GoByte(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xc6,
            "address_prefix": 0x26
        }
    }


class Gridcoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xbe,
            "address_prefix": 0x3e
        }
    }


class GroestlCoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x80,
            "address_prefix": 0x24
        },
        "testnet": {
            "wif_prefix": 0xef,
            "address_prefix": 0x6f
        }
    }


class Gulden(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x62,
            "address_prefix": 0x26
        }
    }


class Helleniccoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xb0,
            "address_prefix": 0x30
        }
    }


class Hempcoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xa8,
            "address_prefix": 0x28
        }
    }


class Horizen(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x80,
            "address_prefix": 0x2089
        }
    }


class Hush(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x80,
            "address_prefix": 0x1cb8
        }
    }


class InsaneCoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x37,
            "address_prefix": 0x66
        }
    }


class InternetOfPeople(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x31,
            "address_prefix": 0x75
        }
    }


class IXCoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x80,
            "address_prefix": 0x8a
        }
    }


class Jumbucks(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xab,
            "address_prefix": 0x2b
        }
    }


class Kobocoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xa3,
            "address_prefix": 0x23
        }
    }


class Komodo(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xbc,
            "address_prefix": 0x3c
        }
    }


class Landcoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xb0,
            "address_prefix": 0x30
        }
    }


class LBRYCredits(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x1c,
            "address_prefix": 0x55
        }
    }


class Linx(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xcb,
            "address_prefix": 0x4b
        }
    }


class Litecoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xb0,
            "address_prefix": 0x30
        },
        "testnet": {
            "wif_prefix": 0xef,
            "address_prefix": 0x6f
        }
    }


class LitecoinCash(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xb0,
            "address_prefix": 0x1c
        }
    }


class LitecoinZ(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x80,
            "address_prefix": 0xab3
        }
    }


class Lkrcoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xb0,
            "address_prefix": 0x30
        }
    }


class Lynx(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xad,
            "address_prefix": 0x2d
        }
    }


class Mazacoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xe0,
            "address_prefix": 0x32
        }
    }


class Megacoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xb2,
            "address_prefix": 0x32
        }
    }


class Minexcoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x80,
            "address_prefix": 0x4b
        }
    }


class Monacoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xb0,
            "address_prefix": 0x32
        }
    }


class Monk(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x37,
            "address_prefix": 0x33
        }
    }


class Myriadcoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xb2,
            "address_prefix": 0x32
        }
    }


class Namecoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x80,
            "address_prefix": 0x34
        }
    }


class Navcoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x96,
            "address_prefix": 0x35
        }
    }


class Neblio(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xb5,
            "address_prefix": 0x35
        }
    }


class Neoscoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xb1,
            "address_prefix": 0x35
        }
    }


class Neurocoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xb5,
            "address_prefix": 0x35
        }
    }


class NewYorkCoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xbc,
            "address_prefix": 0x3c
        }
    }


class NIX(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x80,
            "address_prefix": 0x26
        }
    }


class Novacoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x88,
            "address_prefix": 0x8
        }
    }


class NuBits(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x96,
            "address_prefix": 0x19
        }
    }


class NuShares(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x95,
            "address_prefix": 0x3f
        }
    }


class OKCash(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x3,
            "address_prefix": 0x37
        }
    }


class Omni(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x80,
            "address_prefix": 0x0
        },
        "testnet": {
            "wif_prefix": 0xef,
            "address_prefix": 0x6f
        }
    }


class Onix(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xcb,
            "address_prefix": 0x4b
        }
    }


class Particl(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x6c,
            "address_prefix": 0x38
        }
    }


class Peercoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xb7,
            "address_prefix": 0x37
        }
    }


class Pesobit(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xb7,
            "address_prefix": 0x37
        }
    }


class Phore(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xd4,
            "address_prefix": 0x37
        }
    }


class Pinkcoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x83,
            "address_prefix": 0x3
        }
    }


class Pivx(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xd4,
            "address_prefix": 0x1e
        },
        "testnet": {
            "wif_prefix": 0xef,
            "address_prefix": 0x8b
        }
    }


class PoSWCoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xb7,
            "address_prefix": 0x37
        }
    }


class Potcoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xb7,
            "address_prefix": 0x37
        }
    }


class ProjectCoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x75,
            "address_prefix": 0x37
        }
    }


class Putincoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xb7,
            "address_prefix": 0x37
        }
    }


class Qtum(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x80,
            "address_prefix": 0x3a
        },
        "testnet": {
            "wif_prefix": 0xef,
            "address_prefix": 0x78
        }
    }


class Rapids(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x2e,
            "address_prefix": 0x3d
        }
    }


class Ravencoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x80,
            "address_prefix": 0x3c
        },
        "testnet": {
            "wif_prefix": 0x80,
            "address_prefix": 0x6f
        }
    }


class Reddcoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xbd,
            "address_prefix": 0x3d
        }
    }


class Ripple(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x80,
            "address_prefix": 0x0
        }
    }
    ALPHABET = "rpshnaf39wBUDNEGHJKLM4PQRST7VWXYZ2bcdeCg65jkm8oFqi1tuvAxyz"


class Ritocoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x8b,
            "address_prefix": 0x19
        }
    }


class RSK(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x80,
            "address_prefix": 0x0
        },
        "testnet": {
            "wif_prefix": 0xef,
            "address_prefix": 0x6f
        }
    }


class Rubycoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xbc,
            "address_prefix": 0x3c
        }
    }


class Safecoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xbd,
            "address_prefix": 0x3d
        }
    }


class Saluscoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xbf,
            "address_prefix": 0x3f
        }
    }


class Scribe(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x6e,
            "address_prefix": 0x3c
        }
    }


class ShadowCash(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xbf,
            "address_prefix": 0x3f
        },
        "testnet": {
            "wif_prefix": 0xff,
            "address_prefix": 0x7f
        }
    }


class Slimcoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x46,
            "address_prefix": 0x3f
        },
        "testnet": {
            "wif_prefix": 0x57,
            "address_prefix": 0x6f
        }
    }


class Smileycoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x5,
            "address_prefix": 0x19
        }
    }


class Solarcoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x92,
            "address_prefix": 0x12
        }
    }


class Stash(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xcc,
            "address_prefix": 0x4c
        },
        "testnet": {
            "wif_prefix": 0xef,
            "address_prefix": 0x8c
        }
    }


class Stratis(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xbf,
            "address_prefix": 0x3f
        },
        "testnet": {
            "wif_prefix": 0xbf,
            "address_prefix": 0x41
        }
    }


class Sugarchain(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x80,
            "address_prefix": 0x3f
        },
        "testnet": {
            "wif_prefix": 0xef,
            "address_prefix": 0x42
        }
    }


class Syscoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x80,
            "address_prefix": 0x3f
        }
    }


class ThoughtAI(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x7b,
            "address_prefix": 0x7
        }
    }


class TOACoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xc1,
            "address_prefix": 0x41
        }
    }


class TWINS(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x42,
            "address_prefix": 0x49
        },
        "testnet": {
            "wif_prefix": 0xed,
            "address_prefix": 0x4c
        }
    }


class UltimateSecureCash(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xbf,
            "address_prefix": 0x44
        }
    }


class Unobtanium(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xe0,
            "address_prefix": 0x82
        }
    }


class Vcash(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xc7,
            "address_prefix": 0x47
        }
    }


class Verge(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x9e,
            "address_prefix": 0x1e
        }
    }


class Vertcoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x80,
            "address_prefix": 0x47
        }
    }


class Viacoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xc7,
            "address_prefix": 0x47
        },
        "testnet": {
            "wif_prefix": 0xff,
            "address_prefix": 0x7f
        }
    }


class Vivo(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xc6,
            "address_prefix": 0x46
        }
    }


class Voxels(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xc6,
            "address_prefix": 0x46
        }
    }


class VirtualCash(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xc7,
            "address_prefix": 0x47
        }
    }


class Wagerr(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xc7,
            "address_prefix": 0x49
        }
    }


class Whitecoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xc9,
            "address_prefix": 0x49
        }
    }


class Wincoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xc9,
            "address_prefix": 0x49
        }
    }


class XUEZ(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xd4,
            "address_prefix": 0x4b
        }
    }


class Ycash(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x80,
            "address_prefix": 0x1c28
        }
    }


class Zcash(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x80,
            "address_prefix": 0x1cb8
        },
        "testnet": {
            "wif_prefix": 0xef,
            "address_prefix": 0x1d25
        }
    }


class ZClassic(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x80,
            "address_prefix": 0x1cb8
        }
    }


class Zetacoin(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0xe0,
            "address_prefix": 0x50
        }
    }


class ZooBC(ICryptocurrency):

    NETWORKS = {
        "mainnet": {
            "wif_prefix": 0x80,
            "address_prefix": 0x0
        }
    }



#!/usr/bin/env python3

# Copyright © 2023-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    Union, Tuple, Type
)

from .libs.base58 import (
    encode, decode
)
from .cryptocurrencies import (
    ICryptocurrency, Bitcoin
)
from .const import (
    COMPRESSED_PRIVATE_KEY_PREFIX, WIF_TYPES
)
from .crypto import get_checksum
from .exceptions import (
    WIFError, Secp256k1Error
)
from .utils import (
    get_bytes, integer_to_bytes, bytes_to_string
)


def encode_wif(
    private_key: Union[str, bytes], wif_prefix: int = Bitcoin.NETWORKS["mainnet"]["wif_prefix"]
) -> Tuple[str, str]:
    """
    Encode a private key to Wallet Import Format (WIF).

    :param private_key: The private key to encode, as a 32-byte string or bytes.
    :type private_key: Union[str, bytes]
    :param wif_prefix: The prefix to use for the WIF format (default is Bitcoin mainnet prefix).
    :type wif_prefix: int

    :returns: A tuple containing the WIF and WIF-compressed formats.
    :rtype: Tuple[str, str]
    """

    if len(get_bytes(private_key)) != 32:
        raise Secp256k1Error("Invalid private key length", expected=64, got=len(private_key))

    wif_payload: bytes = (
        integer_to_bytes(wif_prefix) + get_bytes(private_key)
    )
    wif_compressed_payload: bytes = (
        integer_to_bytes(wif_prefix) + get_bytes(private_key) + integer_to_bytes(COMPRESSED_PRIVATE_KEY_PREFIX)
    )
    return (
        encode(wif_payload + get_checksum(wif_payload)),
        encode(wif_compressed_payload + get_checksum(wif_compressed_payload))
    )


def decode_wif(
    wif: str, wif_prefix: int = Bitcoin.NETWORKS["mainnet"]["wif_prefix"]
) -> Tuple[bytes, str, bytes]:
    """
    Decode a Wallet Import Format (WIF) string to a private key.

    :param wif: The WIF string to decode.
    :type wif: str
    :param wif_prefix: The prefix to use for the WIF format (default is Bitcoin mainnet prefix).
    :type wif_prefix: int

    :returns: A tuple containing the private key, the WIF type ('wif' or 'wif-compressed'), and the checksum.
    :rtype: Tuple[bytes, str, bytes]
    """

    raw: bytes = decode(wif)
    if not raw.startswith(integer_to_bytes(wif_prefix)):
        raise WIFError(f"Invalid Wallet Import Format (WIF)")

    prefix_length: int = len(integer_to_bytes(wif_prefix))
    prefix_got: bytes = raw[:prefix_length]
    if integer_to_bytes(wif_prefix) != prefix_got:
        raise WIFError("Invalid WIF prefix", expected=prefix_length, got=prefix_got)

    raw_without_prefix: bytes = raw[prefix_length:]
    checksum: bytes = raw_without_prefix[-1 * 4:]
    private_key: bytes = raw_without_prefix[:-1 * 4]
    wif_type: str = "wif"

    if len(private_key) not in [33, 32]:
        raise WIFError(f"Invalid Wallet Import Format (WIF)")
    elif len(private_key) == 33:
        private_key = private_key[:-len(integer_to_bytes(COMPRESSED_PRIVATE_KEY_PREFIX))]
        wif_type = "wif-compressed"

    return private_key, wif_type, checksum


def private_key_to_wif(
    private_key: Union[str, bytes],
    wif_type: str = "wif-compressed",
    cryptocurrency: Type[ICryptocurrency] = Bitcoin,
    network: str = "mainnet"
) -> str:
    """
    Convert a private key to Wallet Import Format (WIF).

    :param private_key: The private key to convert, as a 32-byte string or bytes.
    :type private_key: Union[str, bytes]
    :param wif_type: The WIF type, either 'wif' or 'wif-compressed' (default is 'wif-compressed').
    :type wif_type: str
    :param cryptocurrency: The cryptocurrency class (default is Bitcoin).
    :type cryptocurrency: Type[ICryptocurrency]
    :param network: The network type (default is 'mainnet').
    :type network: str

    :returns: The private key in WIF format.
    :rtype: str
    """

    if wif_type not in WIF_TYPES:
        raise WIFError("Wrong WIF type", expected=WIF_TYPES, got=wif_type)
    wif, wif_compressed = encode_wif(
        private_key=private_key, wif_prefix=cryptocurrency.NETWORKS[network]["wif_prefix"]
    )
    return wif if wif_type == "wif" else wif_compressed


def wif_to_private_key(
    wif: str, cryptocurrency: Type[ICryptocurrency] = Bitcoin, network: str = "mainnet"
) -> str:
    """
    Convert a Wallet Import Format (WIF) string to a private key.

    :param wif: The WIF string to decode.
    :type wif: str
    :param cryptocurrency: The cryptocurrency class (default is Bitcoin).
    :type cryptocurrency: Type[ICryptocurrency]
    :param network: The network type (default is 'mainnet').
    :type network: str

    :returns: The private key as a string.
    :rtype: str
    """

    return bytes_to_string(decode_wif(
        wif=wif, wif_prefix=cryptocurrency.NETWORKS[network]["wif_prefix"]
    )[0])


def get_wif_type(
    wif: str, cryptocurrency: Type[ICryptocurrency] = Bitcoin, network: str = "mainnet"
) -> str:
    """
    Get the type of Wallet Import Format (WIF) string ('wif' or 'wif-compressed').

    :param wif: The WIF string to inspect.
    :type wif: str
    :param cryptocurrency: The cryptocurrency class (default is Bitcoin).
    :type cryptocurrency: Type[ICryptocurrency]
    :param network: The network type (default is 'mainnet').
    :type network: str

    :returns: The WIF type ('wif' or 'wif-compressed').
    :rtype: str
    """

    return decode_wif(
        wif=wif, wif_prefix=cryptocurrency.NETWORKS[network]["wif_prefix"]
    )[1]


def get_wif_checksum(
    wif: str, cryptocurrency: Type[ICryptocurrency] = Bitcoin, network: str = "mainnet"
) -> str:
    """
    Get the checksum of a Wallet Import Format (WIF) string.

    :param wif: The WIF string to inspect.
    :type wif: str
    :param cryptocurrency: The cryptocurrency class (default is Bitcoin).
    :type cryptocurrency: Type[ICryptocurrency]
    :param network: The network type (default is 'mainnet').
    :type network: str

    :returns: The checksum as a string.
    :rtype: str
    """

    return bytes_to_string(decode_wif(
        wif=wif, wif_prefix=cryptocurrency.NETWORKS[network]["wif_prefix"]
    )[2])



#!/usr/bin/env python3

# Copyright © 2023-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import (
    Any, Union
)

from .libs.base58 import (
    ensure_string, check_encode, check_decode
)
from .secp256k1 import PublicKey
from .cryptocurrencies import Bitcoin
from .crypto import hash160
from .exceptions import AddressError
from .utils import (
    get_bytes, integer_to_bytes, bytes_to_string
)


class P2PKHAddress:
    
    address_prefix: int = Bitcoin.NETWORKS["mainnet"]["address_prefix"]
    alphabet: str = Bitcoin.ALPHABET

    @classmethod
    def encode(cls, public_key: Union[bytes, str, PublicKey], **kwargs: Any) -> str:
        """
        Encode a public key into an address using a specified address prefix and alphabet.

        :param public_key: The public key to encode.
        :type public_key: Union[bytes, str, IPublicKey]
        :param kwargs: Additional keyword arguments.
            - address_prefix: Address prefix for the public key (optional).
            - public_key_type: Type of the public key (optional).
            - alphabet: Custom alphabet for encoding (optional).
        :type kwargs: Any

        :return: The encoded address.
        :rtype: str
        """

        address_prefix: bytes = integer_to_bytes(
            kwargs.get("address_prefix", cls.address_prefix)
        )

        if not isinstance(public_key, PublicKey):
            public_key: PublicKey = PublicKey.from_bytes(get_bytes(public_key))

        public_key_hash: bytes = hash160(
            public_key.raw_compressed()
            if kwargs.get("public_key_type", "compressed") == "compressed" else
            public_key.raw_uncompressed()
        )

        return ensure_string(check_encode(
            (address_prefix + public_key_hash), alphabet=kwargs.get(
                "alphabet", cls.alphabet
            )
        ))

    @classmethod
    def decode(cls, address: str, **kwargs: Any) -> str:
        """
        Decode an address string into a public key hash using a specified address prefix and alphabet.

        :param address: The address string to decode.
        :type address: str
        :param kwargs: Additional keyword arguments.
            - address_prefix: Address prefix for the public key (optional).
            - alphabet: Custom alphabet for decoding (optional).
        :type kwargs: Any

        :return: The decoded public key hash as a string.
        :rtype: str
        """

        address_prefix: bytes = integer_to_bytes(
            kwargs.get("address_prefix", cls.address_prefix)
        )
        address_decode: bytes = check_decode(
            address, alphabet=kwargs.get(
                "alphabet", cls.alphabet
            )
        )

        expected_length: int = 20 + len(address_prefix)
        if len(address_decode) != expected_length:
            raise AddressError(f"Invalid length (expected: {expected_length}, got: {len(address_decode)})")

        prefix_got: bytes = address_decode[:len(address_prefix)]
        if address_prefix != prefix_got:
            raise AddressError(f"Invalid prefix (expected: {address_prefix}, got: {prefix_got})")

        return bytes_to_string(address_decode[len(address_prefix):])



#!/usr/bin/env python3

# Copyright © 2023-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import List
from setuptools import (
    setup, find_packages
)

import importlib.util


# requirements/{name}.txt
def get_requirements(name: str) -> List[str]:
    with open(f"{name}.txt", "r") as requirements:
        return list(map(str.strip, requirements.read().split("\n")))


# README.md
with open("README.md", "r", encoding="utf-8") as readme:
    long_description: str = readme.read()

# hdwallet/info.py
spec = importlib.util.spec_from_file_location(
    "info", "bip38/info.py"
)
info = importlib.util.module_from_spec(spec)
spec.loader.exec_module(info)

setup(
    name=info.__name__,
    version=info.__version__,
    description=info.__description__,
    long_description=long_description,
    long_description_content_type="text/markdown",
    license=info.__license__,
    author=info.__author__,
    author_email=info.__email__,
    url=info.__url__,
    project_urls={
        "Tracker": info.__tracker__,
        "Source": info.__source__,
        "Changelog": info.__changelog__,
        "Documentation": info.__documentation__
    },
    keywords=info.__keywords__,
    python_requires=">=3.9,<4",
    packages=find_packages(exclude=["tests*"]),
    install_requires=get_requirements(name="requirements"),
    include_package_data=True,
    extras_require=dict(
        docs=get_requirements(name="requirements/docs"),
        tests=get_requirements(name="requirements/tests")
    ),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ]
)



#!/usr/bin/env python3

# Copyright © 2023-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import List

from .info import (
    __name__,
    __version__,
    __license__,
    __author__,
    __email__,
    __documentation__,
    __description__,
    __url__,
    __tracker__,
    __keywords__
)
from .bip38 import BIP38

__all__: List[str] = [
    "__name__",
    "__version__",
    "__license__",
    "__author__",
    "__email__",
    "__documentation__",
    "__description__",
    "__url__",
    "__tracker__",
    "__keywords__",
    "BIP38"
]



#!/usr/bin/env python3

# Copyright © 2023-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from typing import Any
from ecdsa import SigningKey
from ecdsa import (
    curves, keys
)

from ..const import PRIVATE_KEY_LENGTH
from ..exceptions import Secp256k1Error
from .public_key import PublicKey


class PrivateKey:

    signing_key: SigningKey

    def __init__(self, signing_key: SigningKey) -> None:
        """
        Initializes an instance with a signing key.

        :param signing_key: The signing key to be used for cryptographic operations.
        :type signing_key: SigningKey
        """

        self.signing_key = signing_key

    @classmethod
    def from_bytes(cls, private_key: bytes) -> "PrivateKey":
        """
        Creates a private key instance from the given bytes.

        :param private_key: The bytes representing the private key.
        :type private_key: bytes

        :return: An instance of PrivateKey corresponding to the given bytes.
        :rtype: PrivateKey
        """

        try:
            return cls(
                SigningKey.from_string(
                    private_key, curve=curves.SECP256k1
                )
            )
        except keys.MalformedPointError as ex:
            raise Secp256k1Error("Invalid private key bytes") from ex

    @staticmethod
    def length() -> int:
        """
        Returns the length of the private key in bytes for  SECP256k1 curve.

        :return: Length of the private key in bytes.
        :rtype: int
        """

        return PRIVATE_KEY_LENGTH

    def underlying_object(self) -> Any:
        """
        Return the underlying signing key object.

        :return: The underlying signing key object.
        :rtype: Any
        """

        return self.signing_key

    def raw(self) -> bytes:
        """
        Return the raw bytes representation of the private key.

        :return: The raw bytes of the private key.
        :rtype: bytes
        """

        return self.signing_key.to_string()

    def public_key(self) -> PublicKey:
        """
        Retrieve the public key associated with this private key instance.

        :return: The public key object.
        :rtype: IPublicKey
        """

        return PublicKey(self.signing_key.get_verifying_key())



#!/usr/bin/env python3

# Copyright © 2023-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from bip38.cryptocurrencies import Bitcoin
from bip38.wif import (
    encode_wif, decode_wif
)


def test_wif(_):

    for network in Bitcoin.NETWORKS.keys():

        uncompressed_wif, compressed_wif = encode_wif(
            private_key=_["wif"]["private_key"], wif_prefix=Bitcoin.NETWORKS[network]["wif_prefix"]
        )
        assert _["wif"]["uncompressed"]["wif"][network] == uncompressed_wif
        assert _["wif"]["compressed"]["wif"][network] == compressed_wif

        private_key, wif_type, checksum = decode_wif(
            compressed_wif, wif_prefix=Bitcoin.NETWORKS[network]["wif_prefix"]
        )
        assert private_key.hex() == _["wif"]["private_key"]
        assert wif_type == _["wif"]["compressed"]["wif_type"]
        assert checksum.hex() == _["wif"]["compressed"]["wif_checksum"][network]

        private_key, wif_type, checksum = decode_wif(
            uncompressed_wif, wif_prefix=Bitcoin.NETWORKS[network]["wif_prefix"]
        )
        assert private_key.hex() == _["wif"]["private_key"]
        assert wif_type == _["wif"]["uncompressed"]["wif_type"]
        assert checksum.hex() == _["wif"]["uncompressed"]["wif_checksum"][network]



#!/usr/bin/env python3

from hashlib import sha256

import six

__base58_alphabet = "123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz"


def string_to_int(data):
    val = 0

    if type(data) == str:
        data = bytearray(data)

    for (i, c) in enumerate(data[::-1]):
        val += (256 ** i) * c
    return val


def ensure_string(data):
    if isinstance(data, six.binary_type):
        return data.decode("utf-8")
    elif not isinstance(data, six.string_types):
        raise ValueError("Invalid value for string")
    return data


def encode(data, alphabet=__base58_alphabet):
    enc = ""
    val = string_to_int(data)
    while val >= len(alphabet):
        val, mod = divmod(val, len(alphabet))
        enc = alphabet[mod] + enc
    if val:
        enc = alphabet[val] + enc

    n = len(data) - len(data.lstrip(b"\0"))
    return alphabet[0] * n + enc


def check_encode(raw, alphabet=__base58_alphabet):
    chk = sha256(sha256(raw).digest()).digest()[:4]
    return encode(raw + chk, alphabet)


def decode(data, alphabet=__base58_alphabet):
    if bytes != str:
        data = bytes(data, "ascii")

    val = 0
    prefix = 0
    for c in data:
        val = (val * len(alphabet)) + alphabet.encode("utf-8").find(c)
        if val == 0:
            prefix += 1

    dec = bytearray()
    while val > 0:
        val, mod = divmod(val, 256)
        dec.append(mod)

    dec.extend(bytearray(prefix))

    return bytes(dec[::-1])


def check_decode(enc, alphabet=__base58_alphabet):
    dec = decode(enc, alphabet)
    raw, chk = dec[:-4], dec[-4:]
    if chk != sha256(sha256(raw).digest()).digest()[:4]:
        raise ValueError("base58 decoding checksum error")
    else:
        return raw
